""" MultiQC module to parse output from QoRTs """


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="QoRTs",
            anchor="qorts",
            href="http://hartleys.github.io/QoRTs/",
            info="is toolkit for analysis, QC and data management of RNA-Seq datasets.",
            doi="10.1186/s12859-015-0670-5",
        )

        # Parse logs
        self.qorts_data = dict()
        for f in self.find_log_files("qorts", filehandles=True):
            self.parse_qorts(f)
            self.add_data_source(f)

        # Remove empty samples
        self.qorts_data = {s: v for s, v in self.qorts_data.items() if len(v) > 0}

        # Filter to strip out ignored sample names
        self.qorts_data = self.ignore_samples(self.qorts_data)

        if len(self.qorts_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.qorts_data)} logs")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.qorts_data, "multiqc_qorts")

        # Make plots
        self.qorts_general_stats()
        self.qorts_alignment_barplot()
        self.qorts_splice_loci_barplot()
        self.qorts_splice_events_barplot()
        self.qorts_strandedness_plot()

    def parse_qorts(self, f):
        s_names = None
        for line in f["f"]:
            s = line.split("\t")
            if s_names is None:
                raw_s_names = s[1:]
                s_names = [self.clean_s_name(s_name, f) for s_name in raw_s_names]
                if len(s_names) <= 2 and raw_s_names[0].endswith("COUNT"):
                    if f["fn"] == "QC.summary.txt":
                        s_names = [self.clean_s_name(os.path.basename(os.path.normpath(f["root"])), f)]
                    else:
                        s_names = [f["s_name"]]
                for s_name in s_names:
                    if s_name in self.qorts_data:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.qorts_data[s_name] = dict()
            else:
                for i, s_name in enumerate(s_names):
                    # Hack to get around Java localisation with commas for decimal places
                    if "," in s[i + 1] and "." not in s[i + 1]:
                        s[i + 1] = s[i + 1].replace(",", ".")
                    self.qorts_data[s_name][s[0]] = float(s[i + 1])
        # Add some extra fields
        for i, s_name in enumerate(s_names):
            if "Genes_Total" in self.qorts_data[s_name] and "Genes_WithNonzeroCounts" in self.qorts_data[s_name]:
                self.qorts_data[s_name]["Genes_PercentWithNonzeroCounts"] = (
                    self.qorts_data[s_name]["Genes_WithNonzeroCounts"] / self.qorts_data[s_name]["Genes_Total"]
                ) * 100.0

    def qorts_general_stats(self):
        """Add columns to the General Statistics table"""
        headers = {
            "Genes_PercentWithNonzeroCounts": {
                "title": "% Genes with Counts",
                "description": "Percent of Genes with Non-Zero Counts",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "NumberOfChromosomesCovered": {
                "title": "Chrs Covered",
                "description": "Number of Chromosomes Covered",
                "format": "{:,.0f}",
            },
        }
        self.general_stats_addcols(self.qorts_data, headers)

    def qorts_alignment_barplot(self):
        """Alignment statistics bar plot"""
        # Specify the order of the different possible categories
        keys = [
            "ReadPairs_UniqueGene_CDS",
            "ReadPairs_UniqueGene_UTR",
            "ReadPairs_AmbigGene",
            "ReadPairs_NoGene_Intron",
            "ReadPairs_NoGene_OneKbFromGene",
            "ReadPairs_NoGene_TenKbFromGene",
            "ReadPairs_NoGene_MiddleOfNowhere",
        ]
        cats = {}
        for k in keys:
            name = k.replace("ReadPairs_", "").replace("_", ": ")
            name = re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", name)
            cats[k] = {"name": name}

        # Config for the plot
        pconfig = {
            "id": "qorts_alignments",
            "title": "QoRTs: Alignment Locations",
            "ylab": "# Read Pairs",
            "cpswitch_counts_label": "Number of Read Pairs",
            "hide_zero_cats": False,
        }

        self.add_section(
            name="Alignments",
            description="This plot displays the rate for which the sample's read-pairs are assigned to the different categories.",
            helptext="""
            The [QoRTs vignette](http://hartleys.github.io/QoRTs/doc/QoRTs-vignette.pdf) describes the categories in this plot as follows:

            * **Unique Gene**: The read-pair overlaps with the exonic segments of one and only one gene. For many
              downstream analyses tools, such as DESeq, DESeq2 and EdgeR, only read-pairs in this category
              are used.
            * **Ambig Gene**: The read-pair overlaps with the exons of more than one gene.
            * **No Gene: Intronic**: The read-pair does not overlap with the exons of any annotated gene, but appears
              in a region that is bridged by an annotated splice junction.
            * **No Gene: One kb From Gene**: The read-pair does not overlap with the exons of any annotated gene, but is
              within 1 kilobase from the nearest annotated gene.
            * **No Gene: Ten kb From Gene**: The read-pair does not overlap with the exons of any annotated gene, but
              is within 10 kilobases from the nearest annotated gene.
            * **No Gene: Middle Of Nowhere**: The read-pair does not overlap with the exons of any annotated gene,
              and is more than 10 kilobases from the nearest annotated gene.

            _What it means and what to look for:_

            Outliers in these plots can indicate biological variations or the presence of large mapping problems.
            They may also suggest the presence of large, highly-expressed, unannotated transcripts or genes.
            """,
            plot=bargraph.plot(self.qorts_data, cats, pconfig),
        )

    def qorts_splice_loci_barplot(self):
        """Make the HighCharts HTML to plot the qorts splice loci"""
        # Specify the order of the different possible categories
        keys = [
            "SpliceLoci_Known_ManyReads",
            "SpliceLoci_Known_FewReads",
            "SpliceLoci_Known_NoReads",
            "SpliceLoci_Novel_ManyReads",
            "SpliceLoci_Novel_FewReads",
        ]
        cats = {}
        for k in keys:
            name = k.replace("SpliceLoci_", "").replace("_", ": ")
            name = re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", name)
            cats[k] = {"name": name}

        # Config for the plot
        pconfig = {
            "id": "qorts_splice_loci",
            "title": "QoRTs: Splice Loci",
            "ylab": "# Splice Loci",
            "cpswitch_counts_label": "Number of Splice Loci",
            "hide_zero_cats": False,
        }

        self.add_section(
            name="Splice Loci",
            description="This plot shows the number of splice junction loci of each type that appear in the sample's reads.",
            helptext="""
            The [QoRTs vignette](http://hartleys.github.io/QoRTs/doc/QoRTs-vignette.pdf) describes the categories in this plot as follows:

            * **Known**: The splice junction locus is found in the supplied transcript annotation gtf file.
            * **Novel**: The splice junction locus is NOT found in the supplied transcript annotation gtf file.
            * **Known: Few reads**: The locus is known, and is only covered by 1-3 read-pairs.
            * **Known: Many reads**: The locus is known, and is covered by 4 or more read-pairs.
            * **Novel: Few reads**: The locus is novel, and is only covered by 1-3 read-pairs.
            * **Novel: Many reads**: The locus is novel, and is covered by 4 or more read-pairs

            _What it means and what to look for:_

            This plot can be used to detect a number of anomalies. For example:
            whether mapping or sequencing artifacts caused a disproportionate discovery of novel splice junctions in
            one sample or batch. It can also be used as an indicator of the comprehensiveness the genome annotation.
            Replicates that are obvious outliers may have sequencing/technical issues causing false detection of splice
            junctions.

            Abnormalities in the splice junction rates are generally a symptom of larger issues which will generally be
            picked up by other metrics. Numerous factors can reduce the efficacy by which aligners map across splice
            junctions, and as such these plots become very important if the intended downstream analyses include
            transcript assembly, transcript deconvolution, differential splicing, or any other form of analysis that in
            some way involves the splice junctions themselves. These plots can be used to assess whether other minor
            abnormalities observed in the other plots are of sufficient severity to impact splice junction mapping and
            thus potentially compromise such analyses.
            """,
            plot=bargraph.plot(self.qorts_data, cats, pconfig),
        )

    def qorts_splice_events_barplot(self):
        """Make the HighCharts HTML to plot the qorts splice events"""
        # Specify the order of the different possible categories
        keys = [
            "SpliceEvents_KnownLociWithManyReads",
            "SpliceEvents_KnownLociWithFewReads",
            "SpliceEvents_NovelLociWithManyReads",
            "SpliceEvents_NovelLociWithFewReads",
        ]
        cats = {}
        for k in keys:
            name = k.replace("SpliceEvents_", "")
            name = re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", name)
            cats[k] = {"name": name}

        # Config for the plot
        pconfig = {
            "id": "qorts_splice_events",
            "title": "QoRTs: Splice Events",
            "ylab": "# Splice Events",
            "cpswitch_counts_label": "Number of Splice Events",
            "hide_zero_cats": False,
        }

        self.add_section(
            name="Splice Events",
            description="This plot shows the number of splice junction events falling into different junction categories.",
            helptext="""
            From the [QoRTs vignette](http://hartleys.github.io/QoRTs/doc/QoRTs-vignette.pdf):

            A splice junction "event" is one instance of a read-pair bridging a splice junction.
            Some reads may contain multiple splice junction events, some may contain none. If a splice junction appears
            on both reads of a read-pair, this is still only counted as a single "event".

            Note that because different samples/runs may have different total read counts and/or library sizes, this function
            is generally not the best for comparing between samples. In general, the event rates per read-pair should be
            used instead.
            This plot is used to detect whether sample-specific or batch effects have a substantial or biased effect on splice
            junction appearance, either due to differences in the original RNA, or due to artifacts that alter the rate at
            which the aligner maps across splice junctions.

            _What it means and what to look for:_

            This plot is useful for identifying mapping and/or annotation issues,
            and can indicate the comprehensiveness the genome annotation. Replicates that are obvious outliers may
            have sequencing/technical issues causing false detection of splice junctions.
            In general, abnormalities in the splice junction rates are generally a symptom of larger issues which will
            often be picked up by other metrics.
            """,
            plot=bargraph.plot(self.qorts_data, cats, pconfig),
        )

    def qorts_strandedness_plot(self):
        """Make a bar plot showing the reads assigned to each strand"""
        # Specify the order of the different possible categories
        keys = [
            "StrandTest_frFirstStrand",
            "StrandTest_frSecondStrand",
            "StrandTest_ambig_genesFountOnBothStrands",
            "StrandTest_ambig_noGenes",
            "StrandTest_ambig_other",
        ]
        cats = {}
        for k in keys:
            name = k.replace("StrandTest_", "").replace("_", " ").replace("ambig", "ambig:")
            name = re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", name)
            cats[k] = {"name": name.title()}

        # Config for the plot
        pconfig = {
            "id": "qorts_strand_test",
            "title": "QoRTs: Strand Test",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
            "cpswitch_c_active": False,
        }

        self.add_section(
            name="Strandedness",
            description="This plot shows the rate at which reads appear to follow different library-type strandedness rules.",
            helptext="""
            From the [QoRTs vignette](http://hartleys.github.io/QoRTs/doc/QoRTs-vignette.pdf):

            This plot is used to detect whether your data is indeed stranded, and whether you are using the correct
            stranded data library type option. For unstranded libraries, one would expect close to 50-50
            First Strand - Second Strand. For stranded libraries, all points should fall closer to 99% one or the other.
            """,
            plot=bargraph.plot(self.qorts_data, cats, pconfig),
        )



#!/usr/bin/env python

""" MultiQC lint helpers. Simple additional tests to run when
--strict is specified (outside scope of normal functions) """


import glob
import os

import yaml

from multiqc.utils import config, report

logger = config.logger


def run_tests():
    """Run all lint tests"""
    if config.strict:
        check_mods_docs_readme()


def lint_error(msg):
    """Add a lint error to the report"""
    logger.error(msg)
    report.lint_errors.append(msg)


def check_mods_docs_readme():
    """Check that all modules are listed in the YAML index
    at the top of docs/README.md"""

    docs_mods = []

    docs_dir = os.path.join(os.path.dirname(config.MULTIQC_DIR), "docs", "modules")
    if not os.path.isdir(docs_dir) and os.environ.get("GITHUB_WORKSPACE"):
        docs_dir = os.path.join(os.environ.get("GITHUB_WORKSPACE"), "docs", "modules")
    if not os.path.isdir(docs_dir):
        logger.warning(f"Can't check docs readmes in lint test as directory doesn't exist: {docs_dir}")
        return None

    for fn in glob.glob(os.path.join(docs_dir, "*.md")):
        docs_mods.append(os.path.basename(fn)[:-3])
    logger.info(f"Checking docs readmes in '{docs_dir}' as --strict specified")

    # Check that installed modules are listed in docs/modules
    for m in config.avail_modules.keys():
        if m not in docs_mods and m != "custom_content":
            lint_error(f"LINT: Module '{m}' found in installed modules, but not docs/modules")

    # Check that modules in docs/modules are installed
    for m in docs_mods:
        if m not in config.avail_modules.keys() and m != "custom_content":
            lint_error(f"LINT: Module '{m}' found in docs/modules, but not installed modules")

    # Check that all modules have a YAML header conforming to the required structure
    for fn in glob.glob(os.path.join(docs_dir, "*.md")):
        with open(fn) as fh:
            # Load the YAML header from the markdown file. YAML header should be placed between --- and --- in the beginning of the file
            try:
                header = fh.read().split("---")[1]
            except IndexError:
                lint_error(f"LINT: '{fn}' doesn't have a YAML header between '---'")
                continue
            try:
                header = yaml.safe_load(header)
            except yaml.YAMLError as e:
                lint_error(f"LINT: '{fn}' contains an incorrectly formatted YAML header: {e}")
                continue
            if header is None:
                lint_error(f"LINT: '{fn}' contains an empty YAML header")
                continue
            req_fields = ["name", "url", "description"]
            for field in req_fields:
                if field not in header:
                    lint_error(
                        f"LINT: the YAML header in '{fn}' does not have a '{field}' field. Required fields: {req_fields}"
                    )



""" MultiQC module to parse output from WhatsHap """

import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


ALL_CHROM = "ALL"


class MultiqcModule(BaseMultiqcModule):
    """
    WhatsHap module class, parses WhatsHap output.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="WhatsHap",
            anchor="whatshap",
            href="https://whatshap.readthedocs.io/",
            info="""is a program for phasing genomic variants using DNA sequencing
                    reads, also called read-based phasing or haplotype assembly. It
                    is especially suitable for long reads, but also works well with
                    short reads.
                    """,
            doi="10.1101/085050",
        )

        # Store the whatshap stats results
        self.whatshap_stats = dict()

        # Iterate over all files we found
        for f in self.find_log_files("whatshap/stats", filehandles=True):
            sample, data = self.parse_whatshap_stats(f)
            self.whatshap_stats[sample] = data
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.whatshap_stats = self.ignore_samples(self.whatshap_stats)

        if not self.whatshap_stats:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.whatshap_stats, "multiqc_whatshap_stats")

        # Add whatshap stats to general statistics
        self.whatshap_add_general_stats()

        # Create bargraph with total phased bp
        self.add_bargraph_total_phased()

        # Add a table with more extensive whatshap stats not shown in the
        # general statistics table
        self.add_stats_table()

    def parse_whatshap_stats(self, logfile):
        """Parse WhatsHap stats file"""

        def parse_numeric_values(data):
            """
            Process the data to convert it to the proper format

            In practice, this only involves the following changes:
                - Convert all numeric values to integers or floats
                - Replace 'nan' values with zero
            """
            for key, value in data.items():
                # Try to make the value an int
                try:
                    data[key] = int(data[key])
                except ValueError:
                    try:
                        data[key] = float(data[key])
                    except ValueError:
                        pass

                # Replace 'nan' with zero
                if value == "nan":
                    data[key] = None

        file_content = logfile["f"]
        # This will later be replaced by the sample name from the file, unless
        # we are parsing an empty file
        sample = logfile["s_name"]
        # Get the header and remove the "#" character from the first field
        header = next(file_content).strip().split("\t")
        header[0] = header[0].lstrip("#")

        # The results from this log file, a dictionary for every chromosome
        results = defaultdict(dict)
        # Parse the lines that have the data
        for line in file_content:
            fields = line.strip().split("\t")
            if len(header) != len(fields):
                continue

            data = {field: value for field, value in zip(header, fields)}
            if "chromosome" not in data or "sample" not in data:
                continue

            # Remove the sample and chromosome from the data
            sample = str(data.pop("sample"))
            chromosome = data.pop("chromosome")

            # Process the remaining data fields
            parse_numeric_values(data)

            # Insert the current line under chromosome
            results[chromosome] = data

        # If we were parsing an empty file, we return dummy data (basically all
        # zero) so the sample statistics are consistent, which eases parsing
        # later on. WhatsHap uses the fictional 'ALL' chromosome to store the
        # summary for all results, so we re-use that to store all zero's
        if not results:
            results = dict()
            results[ALL_CHROM] = defaultdict(int)

        # If ALL chromosome is not present, and there are multiple contigs available,
        # e.g. in case of a truncated input file, we summarize available contigs
        # as a fictional 'ALL' chromosome.
        if ALL_CHROM not in results and len(results) > 1:
            log.warning(
                f"Could not find the 'ALL' chromosome for {sample}. Make sure that the input is not truncated. MultiQC will summarize information for all available chromosomes instead."
            )
            sum_fields = {
                "variants",
                "phased",
                "unphased",
                "singletons",
                "blocks",
                "bp_per_block_sum",
                "heterozygous_variants",
                "heterozygous_snvs",
                "phased_snvs",
            }
            avg_fields = {
                "bp_per_block_avg",
                "variant_per_block_avg",
                "bp_per_block_avg",
            }
            results[ALL_CHROM] = defaultdict(int)
            for chrom in results:
                if chrom == ALL_CHROM:
                    continue
                for field in results[chrom]:
                    val = results[chrom][field]
                    if val:
                        if field in sum_fields:
                            results[ALL_CHROM][field] += int(results[chrom][field])
                        elif field in avg_fields:
                            results[ALL_CHROM][field] += float(results[chrom][field]) * results[chrom]["blocks"]
            for field in avg_fields:
                results[ALL_CHROM][field] /= results[ALL_CHROM]["blocks"]

        # Calculate the fraction of heterozygous variants that were phased
        for chrom, data in results.items():
            try:
                frac_het_phased = data["phased"] / data["heterozygous_variants"]
            except ZeroDivisionError:
                frac_het_phased = 0
            results[chrom]["frac_het_phased"] = frac_het_phased

        # Clean the sample name
        sample = self.clean_s_name(sample, logfile["root"])

        return sample, results

    @staticmethod
    def get_summary_field(sample_stats):
        """
        Return the summary field for sample_stats

        If there is only a single chromosome, we use that as the summary field.
        If there are multiple chromosomes, whatshap adds the 'ALL' field which
        contains the summary
        """
        # If there is only a single chromosome we use that as the summary
        # filed. Otherwise, we use the 'ALL' chromosome (it only gets added
        # by WhatsHap stats when there are multiple chromosomes).
        if len(sample_stats) == 1:
            summary_field = list(sample_stats)[0]
        else:
            summary_field = ALL_CHROM
        return summary_field

    def whatshap_add_general_stats(self):
        """Add WhatsHap stats to the general statistics table"""

        # Store the configuration for each column, this is also used to
        # determine which columns to add to the general statistics table
        # https://whatshap.readthedocs.io/en/latest/guide.html#the-tsv-statistics-format
        general_stats_headers = {
            "frac_het_phased": {
                "id": "perc_het_phased",
                "title": "% Phased Variants",
                "description": """Fraction of heterozygous variants
                                          that could be phased.
                                        """,
                "format": "{:,.0f}",
                "suffix": "%",
                "min": 0,
                "max": 100,
                "modify": lambda x: x * 100,
                "scale": "GnBu",
                "hidden": False,
            },
            "bp_per_block_avg": {
                "id": "bp_per_block_avg",
                "title": "Avg bp per Block",
                "description": """Description of the distribution of non-singleton
                                        block lengths, where the length of a block is the
                                        number of base pairs it covers minus 1. That is, a
                                        block with two variants at positions 2 and 5 has
                                        length 3.""",
                "format": "{:,.0f}",
                "scale": "Greens",
                "hidden": False,
            },
            "block_n50": {
                "id": "block_n50",
                "title": "NG50",
                "description": """The NG50 value of the distribution of the block
                                        lengths. Interleaved blocks are cut in order to
                                        avoid artificially inflating this value.""",
                "format": "{:,.0f}",
                "scale": "Blues",
                "hidden": True,
            },
        }

        general = dict()
        for sample, sample_stats in self.whatshap_stats.items():
            # Get the summary field
            summary_field = self.get_summary_field(sample_stats)

            # The summary data we are adding to the general statistics
            summary_data = sample_stats[summary_field]

            general_stats = {field: summary_data[field] for field in general_stats_headers}

            general[sample] = general_stats

        self.general_stats_addcols(general, general_stats_headers)

    def add_bargraph_total_phased(self):
        """Add a bargraph of the total number of phased base pairs"""
        pdata = dict()
        for sample, sample_stats in self.whatshap_stats.items():
            # Get the summary field
            summary_field = self.get_summary_field(sample_stats)
            pdata[sample] = {"Phased Base Pairs": sample_stats[summary_field]["bp_per_block_sum"]}

        configuration = {
            "id": "multiqc_whatshap_phased_bp_plot",
            "title": "WhatsHap: Phased Basepairs per Sample",
            "anchor": "multiqc_whatshap_phased_bp",
            "ylab": "Base Pairs",
            "cpswitch": False,
            "tt_percentages": False,
        }

        keys = {"Phased Base Pairs": {"name": "Phased Base Pairs"}}

        # If the Phased Base Pairs is zero for all samples, we do not add the
        # WhatsHap section
        for sample, values in pdata.items():
            if values["Phased Base Pairs"]:
                # If we found one sample with data, we break out of the loop
                # and add the WhatsHap section
                break
        # If we completed the for loop, there is no data, so we return without
        # adding the WhatsHap section
        else:
            return

        self.add_section(
            name="Phased Basepairs per Sample",
            anchor="multiqc_whatshap_phased_bp",
            description="""
                    This plot show the total number of phased base pairs for
                    each sample.
                """,
            plot=bargraph.plot(pdata, keys, configuration),
        )

    def add_stats_table(self):
        """Add WhatsHap stats to the general statistics table"""

        # Store the configuration for each column, this is also used to
        # determine which columns to add to the general statistics table
        # https://whatshap.readthedocs.io/en/latest/guide.html#the-tsv-statistics-format
        stats_headers = {
            "variants": {
                "id": "variants",
                "title": "Input Variants",
                "description": """Number of biallelic variants in the input VCF, but
                                        excluding any non-SNV variants if --only-snvs was
                                        used""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "heterozygous_variants": {
                "id": "heterozygous_variants",
                "title": "Heterozygous Variants",
                "description": """The number of biallelic, heterozygous variants in
                                        the input VCF. This is a subset of Input Variants.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "heterozygous_snvs": {
                "id": "heterozygous_snvs",
                "title": "Heterozygous SNVs",
                "description": """The number of biallelic, heterozygous SNVs in the
                                        input VCF. This is a subset of Heterozygous
                                        Variants.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "unphased": {
                "id": "unphased",
                "title": "Unphased Variants",
                "description": """The number of biallelic, heterozygous variants that
                                        are not marked as phased in the input VCF. This
                                        is a subset of heterozygous_variants.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "phased": {
                "id": "phased",
                "title": "Phased Variants",
                "description": """The number of biallelic, heterozygous variants that
                                        are marked as phased in the input VCF. This is
                                        a subset of heterozygous_variants. Also, phased +
                                        unphased + singletons = heterozygous_variants.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "phased_snvs": {
                "id": "phased_snvs",
                "title": "Phased SNVs",
                "description": """The number of biallelic, heterozygous SNVs that are
                                        marked as phased in the input VCF. This is a subset
                                        of phased.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "blocks": {
                "id": "blocks",
                "title": "Blocks",
                "description": "The total number of phase sets/blocks.",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "singletons": {
                "id": "singletons",
                "title": "Singletons",
                "description": "The number of blocks that contain exactly one variant.",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "bp_per_block_sum": {
                "id": "bp_per_block_sum",
                "title": "Total Phased bp",
                "description": """The sum of the lengths of all non-singleton
                                        blocks, where the length of a block is the
                                        number of base pairs it covers minus 1. That is, a
                                        block with two variants at positions 2 and 5 has
                                        length 3.""",
                "format": "{:,.0f}",
                "hidden": False,
            },
            "variant_per_block_avg": {
                "id": "variant_per_block_avg",
                "title": "Avg Variants per Block",
                "description": """Description of the distribution of non-singleton
                                        block sizes, where the size of a block is the number
                                        of variants it contains. Number of biallelic
                                        variants in the input VCF, but excluding any non-SNV
                                        variants if --only-snvs was used.""",
                "hidden": False,
            },
            "frac_het_phased": {
                "id": "perc_het_phased",
                "title": "% Phased Variants",
                "description": """Fraction of heterozygous variants
                                          that could be phased.
                                        """,
                "format": "{:,.0f}",
                "suffix": "%",
                "min": 0,
                "max": 100,
                "modify": lambda x: x * 100,
                "hidden": True,
            },
            "bp_per_block_avg": {
                "id": "bp_per_block_avg",
                "title": "Avg bp per Block",
                "description": """Description of the distribution of non-singleton
                                        block lengths, where the length of a block is the
                                        number of base pairs it covers minus 1. That is, a
                                        block with two variants at positions 2 and 5 has
                                        length 3.""",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "block_n50": {
                "id": "block_n50",
                "title": "NG50",
                "description": """The NG50 value of the distribution of the block
                                        lengths. Interleaved blocks are cut in order to
                                        avoid artificially inflating this value.""",
                "format": "{:,.0f}",
                "hidden": True,
            },
        }

        general = dict()
        for sample, sample_stats in self.whatshap_stats.items():
            # Get the summary field
            summary_field = self.get_summary_field(sample_stats)

            # The summary data we are adding to the general statistics
            summary_data = sample_stats[summary_field]

            stats = {field: summary_data[field] for field in stats_headers}

            general[sample] = stats

        self.add_section(
            name="WhatsHap statistics",
            anchor="whatshap-table",
            plot=table.plot(
                general,
                stats_headers,
                {
                    "namespace": "WhatsHap",
                    "id": "whatshap-stats-table",
                    "title": "WhatsHap Statistics",
                },
            ),
        )



""" MultiQC module to parse output from GATK """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import the GATK submodules
# import varianteval
from .analyze_saturation_mutagenesis import AnalyzeSaturationMutagenesisMixin
from .base_recalibrator import BaseRecalibratorMixin
from .varianteval import VariantEvalMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, AnalyzeSaturationMutagenesisMixin, BaseRecalibratorMixin, VariantEvalMixin):
    """GATK has a number of different commands and outputs.
    This MultiQC module supports some but not all. The code for
    each script is split into its own file and adds a section to
    the module output if logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="GATK",
            anchor="gatk",
            target="GATK",
            href="https://www.broadinstitute.org/gatk/",
            info=" is a toolkit offering a wide variety of tools with a primary focus on variant discovery and genotyping.",
            doi=["10.1101/201178", "10.1002/0471250953.bi1110s43", "10.1038/ng.806", "10.1101/gr.107524.110"],
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = {}
        self.general_stats_data = {}

        # Call submodule functions
        n_reports_found = 0
        n_reports_found += self.parse_gatk_analyze_saturation_mutagenesis()
        n_reports_found += self.parse_gatk_base_recalibrator()
        n_reports_found += self.parse_gatk_varianteval()

        # Exit if we didn't find anything
        if n_reports_found == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)

    def parse_report(self, lines, table_names):
        """Parse a GATK report https://software.broadinstitute.org/gatk/documentation/article.php?id=1244

        Only GATTable entries are parsed.  Tables are returned as a dict of tables.
        Each table is a dict of arrays, where names correspond to column names, and arrays
        correspond to column values.

        Args:
            lines (file handle): an iterable over the lines of a GATK report.
            table_names (dict): a dict with keys that are GATK report table names
                (e.g. "#:GATKTable:Quantized:Quality quantization map"), and values that are the
                keys in the returned dict.

        Returns:
            {
                table_1:
                    {
                        col_1: [ val_1, val_2, ... ]
                        col_2: [ val_1, val_2, ... ]
                        ...
                    }
                table_2:
                    ...
            }
        """

        report = dict()
        lines = (line for line in lines)
        for line in lines:
            line = line.rstrip()
            if line in table_names.keys():
                report[table_names[line]] = self.parse_gatk_report_table(lines)
        return report

    @staticmethod
    def parse_gatk_report_table(lines):
        headers = next(lines).rstrip().split()
        table = {h: [] for h in headers}
        for line in lines:
            line = line.rstrip()

            # testing to see if we have reached the end of a table in a GATKReport
            if line == "":
                break

            for index, value in enumerate(line.split()):
                table[headers[index]].append(value)
        return table



#!/usr/bin/env python

""" MultiQC module to parse similarity matrix output by sourmash compare """

import logging
import os
import re

import numpy

from multiqc.plots import heatmap

# Initialise the logger
log = logging.getLogger(__name__)


class CompareMixin:
    def parse_compare(self):
        """
        Modeled after vcftools relatedness2 module, which also has many samples represented in the parsed file.
        """
        matrices = {}

        for f in self.find_log_files("sourmash/compare", filehandles=True):
            labels = [x.strip() for x in f["f"]]
            if labels:
                matrix_path = re.sub(".labels.txt", "", f["f"].name)
                if not os.path.exists(matrix_path):
                    log.warning(
                        f"Found a 'labels' file expected by Sourmash: '{f['f'].name}', "
                        f"however, could not find a accompanying matrix binary file "
                        f"'{matrix_path}'. So assuming that wasn't a Sourmash result"
                    )
                    continue
                with open(matrix_path, "rb") as fh:
                    matrix = numpy.load(fh)
                # Note that "s_name" here is not a sample name, but the name of the
                # input file, that contains a comparison matrix across multiple samples.
                matrices[f["s_name"]] = (labels, matrix.tolist())
                self.add_data_source(f, section="compare")

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, f["s_name"])

        matrices = self.ignore_samples(matrices)
        if len(matrices) == 0:
            return 0

        log.info(f"Found {len(matrices)} valid compare results")

        self.write_data_file(matrices, "sourmash_compare")

        helptext = """
        Sourmash `compare` calculates the similarity score between samples. A higher score indicates a higher degree of
        similarity, up to a maximum of 1. Samples are clustered by similarity on each axis, and specific IDs can be
        found in the graph with the Highlight tab.
        """

        for name, (labels, data) in matrices.items():
            # Note that "name" here is not a sample name, but the name of the input file,
            # that contains a comparison matrix across multiple samples.
            id = name.lower().strip().replace(" ", "-").replace(".labels.txt", "")
            self.add_section(
                name=f"Sample similarity (<code>{name}</code>)",
                anchor=f"sourmash-compare-{id}",
                description=f"Heatmap of similarity values from the output of `sourmash compare` run on <code>{name}</code>",
                helptext=helptext,
                plot=heatmap.plot(
                    data,
                    xcats=labels,
                    ycats=labels,
                    pconfig={
                        "id": f"sourmash-compare-heatmap-{id}",
                        "title": "Sourmash: Compare",
                        "square": True,
                        "decimalPlaces": 7,
                    },
                ),
            )

        return len(matrices)



""" MultiQC submodule to parse output from fgbio ErrorRateByReadPosition """


from multiqc.plots import linegraph
from multiqc.utils.util_functions import strtobool


def parse_reports(self):
    """Parser metric files for fgbio ErrorRateByReadPosition.

    Stores the per-read-per-position metrics into a data file and adds a section
    with a per-sample plot.
    """
    linegraph_keys = [
        "error_rate",
        "a_to_c_error_rate",
        "a_to_g_error_rate",
        "a_to_t_error_rate",
        "c_to_a_error_rate",
        "c_to_g_error_rate",
        "c_to_t_error_rate",
    ]
    non_collapsed_keys = [
        "g_to_a_error_rate",
        "g_to_c_error_rate",
        "g_to_t_error_rate",
        "t_to_a_error_rate",
        "t_to_c_error_rate",
        "t_to_g_error_rate",
    ]

    # slurp in all the data
    all_data = dict()
    error_rates = dict()
    y_max = 0.01  # default to 1%
    collapse = True  # same as the `--collapse` option on `ErrorRateByReadPosition`
    is_new_format = False  # Test if this is new or old format
    for f in self.find_log_files("fgbio/errorratebyreadposition", filehandles=True):
        self.add_data_source(f, section="Error rate by read position")

        fh = f["f"]
        header = fh.readline().rstrip("\r\n").split("\t")
        if not header or header[0] != "read_number":
            continue

        # Check if this is the new style
        if "collapsed" in header:
            is_new_format = True

        # slurp in the data for this sample
        s_name = f["s_name"]
        s_data = dict()
        bases_total = 0
        errors = 0
        for line in fh:
            fields = line.rstrip("\r\n").split("\t")
            assert len(fields) == len(header), f"Missing fields in line: `{line}`"
            fields[1:4] = [int(field) for field in fields[1:4]]
            fields[4:11] = [float(field) for field in fields[4:11]]
            if is_new_format:
                # Check if collapse was true or false
                fields[-1] = strtobool(fields[-1])
                collapse = fields[-1]
                if not collapse:
                    # Substitutions types were not collapsed, parse them
                    fields[11:-1] = [float(field) for field in fields[11:-1]]

            row_data = dict(zip(header, fields))
            read_number = row_data["read_number"]
            position = row_data["position"]
            if read_number not in s_data:
                s_data[read_number] = dict()
            s_data[read_number][position] = row_data

            for key in linegraph_keys + (non_collapsed_keys if not collapse else []):
                y_max = max(y_max, row_data[key])
            bases_total += row_data["bases_total"]
            errors += row_data["errors"]

        if s_data:
            all_data[s_name] = s_data
            error_rate = 0.0 if bases_total == 0 else errors / float(bases_total)
            error_rates[s_name] = {"error_rate": error_rate}

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None, f["s_name"])

    # ignore samples
    all_data = self.ignore_samples(all_data)

    # if no data, then do nothing
    if not all_data:
        return 0

    # Write parsed data to a file
    self.write_data_file(all_data, "multiqc_fgbio_ErrorRateByReadPosition_per_position")
    self.write_data_file(error_rates, "multiqc_fgbio_ErrorRateByReadPosition_total")

    # Plot the data and add section
    pconfig = {
        "id": "fgbio_ErrorRateByReadPosition",
        "title": "Fgbio: Error Rate by Read Position",
        "ylab": "Error rate",
        "xlab": "Read Position",
        "xDecimals": False,
        "tt_label": "<b>read position {point.x}</b>: {point.y:.2f} %",
        "ymax": y_max,
        "ymin": 0,
        "data_labels": [
            {"name": "Error Rate", "ylab": "Error Rate"},
            {"name": "A > C", "ylab": "A to C error rate"},
            {"name": "A > G", "ylab": "A to G error rate"},
            {"name": "A > T", "ylab": "A to T error rate"},
            {"name": "C > A", "ylab": "C to A error rate"},
            {"name": "C > G", "ylab": "C to G error rate"},
            {"name": "C > T", "ylab": "C to T error rate"},
        ],
    }

    uncollapsed_labels = [
        {"name": "G > A", "ylab": "G to A error rate"},
        {"name": "G > C", "ylab": "G to C error rate"},
        {"name": "G > T", "ylab": "G to T error rate"},
        {"name": "T > A", "ylab": "T to A error rate"},
        {"name": "T > C", "ylab": "T to C error rate"},
        {"name": "T > G", "ylab": "T to G error rate"},
    ]

    if not collapse:
        pconfig["data_labels"] += uncollapsed_labels

    keys = linegraph_keys + (non_collapsed_keys if not collapse else [])

    # Build a list of linegraphs
    linegraph_data = [{} for _ in keys]
    for s_name, s_data in all_data.items():
        for read_number, read_data in s_data.items():
            s_name_with_read = "%s_R%d" % (s_name, int(read_number))
            for index, lg in enumerate(linegraph_data):
                lg[s_name_with_read] = dict((d["position"], d[keys[index]]) for d in read_data.values())

    # add a section for the plot
    self.add_section(
        name="Error Rate by Read Position",
        anchor="fgbio-error-rate-by-read-position",
        description="Error rate by read position. Plot tabs show the error rates for specific substitution types. `--collapse={}`".format(
            collapse
        ),
        helptext="""
        The error rate by read position. If `collapsed` was `true`, then complementary
        substitutions were grouped together into the first 6 error rates.
        e.g. `T>G` substitutions are reported as `A>C`. Otherwise, all 12 substitution
        rates are reported.


        The following are reads / bases are excluded from the analysis:

        * Unmapped reads
        * Reads marked as failing vendor quality
        * Reads marked as duplicates (unless `--include-duplicates` was specified)
        * Secondary and supplemental records
        * Soft-clipped bases in records
        * Reads with MAPQ < `--min-mapping-quality` (default: `20`)
        * Bases with base quality < `--min-base-quality` (default: `0`)
        * Bases where either the read base or the reference base is non-ACGT
        """,
        plot=linegraph.plot(linegraph_data, pconfig),
    )

    # Add to general stats table
    headers = {
        "error_rate": {
            "title": "% Error",
            "description": "Percent error across all read positions",
            "min": 0,
            "max": 100.0,
            "scale": "RdYlGn-rev",
            "suffix": "%",
            "format": "{:,.2f}",
            "modify": lambda x: 100.0 * x,
        }
    }
    self.general_stats_addcols(error_rates, headers)

    return len(all_data)



import logging
from multiqc.plots import linegraph

log = logging.getLogger(__name__)


def parse_reports(self):
    # To store the summary data
    self.neighbours = dict()

    # Parse the output files
    parse_log_files(self)

    # Remove filtered samples
    self.neighbours = self.ignore_samples(self.neighbours)

    if self.neighbours:
        self.write_data_file(self.neighbours, "multiqc_humid_neighbours")
        add_to_humid_section(self)


def parse_log_files(self):
    for f in self.find_log_files("humid/neighbours"):
        # There is no sample name in the log, so we use the root of the
        # file as sample name (since the filename is always stats.dat
        s_name = self.clean_s_name(f["root"], f)

        # process the file content
        d = {}
        for line in f["contents_lines"]:
            nr_neighbours, count = line.strip("\n").split(" ")
            d[int(nr_neighbours)] = int(count)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Got this far, data must be good
        if s_name in self.neighbours:
            log.debug("Duplicate sample name found! Overwriting: {s_name}")
        self.neighbours[s_name] = d
        self.add_data_source(f, s_name)


def add_to_humid_section(self):
    # Figure configuration
    plot_config = {
        "id": "humid-neighbours",
        "title": "HUMID: Neighbour statistics",
        "ylab": "Number of sequences",
        "xlab": "Number of neighbours",
        "logswitch": True,
        "logswitch_active": True,
        "hide_zero_cats": False,
    }
    self.add_section(
        name="Neighbour statistics",
        anchor="humid-neighbour-section",
        description="""
            Neighbour statistics per sample. For every unique read in the input data,
            this shows the number of neighbours in the graph.
            """,
        plot=linegraph.plot(self.neighbours, plot_config),
    )



#!/usr/bin/env python

""" MultiQC module to parse output from bracken """

import logging

from multiqc.modules.kraken import MultiqcModule as KrakenModule

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(KrakenModule):
    """Bracken module"""

    def __init__(self):
        # Initialise the parent object
        super().__init__(
            name="Bracken",
            anchor="bracken",
            href="https://ccb.jhu.edu/software/bracken/",
            info="is a highly accurate statistical method that computes the abundance of species in DNA sequences from a metagenomics sample.",
            doi="10.7717/peerj-cs.104",
            sp_key="bracken",
        )



""" MultiQC module to parse output from KAT """
import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="K-mer Analysis Toolkit",
            anchor="kat",
            href="https://github.com/TGAC/KAT",
            info="is a toolkit for analysing sequencing data via its k-mer spectra.",
            doi="10.1093/bioinformatics/btw663",
        )

        # Find and load any KAT dist analysis reports
        self.kat_data = dict()
        for f in self.find_log_files("kat"):
            s_name = self.clean_s_name(f["s_name"].replace(".dist_analysis", ""), f)
            content = json.loads(f["f"])
            self.kat_data[s_name] = self.parse_kat_report(content)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.kat_data = self.ignore_samples(self.kat_data)

        if len(self.kat_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.kat_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.kat_data, "multiqc_kat")

        headers = {
            "kmer_peaks": {
                "title": "# of Kmer Peaks",
                "description": "Number of peaks identified in the K-mer spectra",
                "scale": False,
                "format": "{:,.0f}",
            },
            "gc_peaks": {
                "title": "# of GC Peaks",
                "description": "Number of peaks identified in the GC distribution",
                "scale": False,
                "format": "{:,.0f}",
            },
            "est_genome_size": {
                "title": "Est. genome Size",
                "description": "Estimated Genome Size based on K-mer spectra",
                "scale": "BuPu",
                "format": "{:,.0f}",
            },
            "mean_kmer_freq": {
                "title": "Mean K-mer Freq.",
                "description": "Mean K-mer Frequency, provides an estimate of sequencing coverage",
                "scale": "Greens",
                "format": "{:,.0f}",
                "suffix": "x",
            },
        }

        kat_config = {
            "namespace": "KAT",
            "id": "kat-table",
            "title": "KAT Distribution Analysis",
        }

        # Basic Stats Table
        self.add_section(
            name="KAT Distribution Analysis",
            anchor="kat-first",
            description="Table showing k-mer coverage distributions and if available GC distributions",
            helptext="This table can give a quick idea of potential contaminants that can be identified via unexpected numbers of k-mer or gc peaks in the data",
            plot=table.plot(self.kat_data, headers, kat_config),
        )

    def parse_kat_report(self, content):
        table_data = {}
        if "gc" in content and "coverage" in content:
            # GCP
            table_data["kmer_peaks"] = content["coverage"]["nb_peaks"]
            table_data["mean_kmer_freq"] = content["coverage"]["mean_freq"]
            table_data["est_genome_size"] = content["coverage"]["est_genome_size"]
            table_data["gc_peaks"] = content["gc"]["nb_peaks"]
        elif "main_dist" in content:
            # Spectra CN
            table_data["kmer_peaks"] = content["main_dist"]["nb_peaks"]
            table_data["mean_kmer_freq"] = content["main_dist"]["mean_freq"]
            table_data["est_genome_size"] = content["main_dist"]["est_genome_size"]
            table_data["gc_peaks"] = 0
        elif "k" in content:
            # Hist
            table_data["kmer_peaks"] = content["nb_peaks"]
            table_data["mean_kmer_freq"] = content["mean_freq"]
            table_data["est_genome_size"] = content["est_genome_size"]
            table_data["gc_peaks"] = 0
        else:
            log.error("Unexpected JSON configuration")

        return table_data



import dataclasses
import logging
from typing import Dict, List, Union, Any, Optional
import copy

import math
import numpy as np
import plotly.graph_objects as go

from multiqc.utils import config, util_functions
from multiqc.plots.table_object import DataTable
from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset
from multiqc.plots.plotly.table import make_table

logger = logging.getLogger(__name__)


def plot(dt: DataTable, show_table_by_default=False) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    """
    p = ViolinPlot.from_dt(dt, show_table_by_default)

    from multiqc.utils import report

    return p.add_to_report(report)


class ViolinPlot(Plot):
    VIOLIN_HEIGHT = 70  # single violin height
    EXTRA_HEIGHT = 63  # extra space for the title and footer

    @dataclasses.dataclass
    class Dataset(BaseDataset):
        metrics: List[str]
        header_by_metric: Dict[str, Dict[str, Any]]
        violin_values_by_sample_by_metric: Dict[str, Dict[str, Union[List[int], List[float], List[str]]]]
        scatter_values_by_sample_by_metric: Dict[str, Dict[str, Union[List[int], List[float], List[str]]]]
        all_samples: List[str]  # unique list of all samples in this dataset
        scatter_trace_params: Dict[str, Any]

        @staticmethod
        def create(
            dataset: BaseDataset,
            values_by_sample_by_metric: Dict[str, Dict[str, Union[List[int], List[float], List[str]]]],
            header_by_metric: Dict[str, Dict[str, Any]],
        ) -> "ViolinPlot.Dataset":
            ds = ViolinPlot.Dataset(
                **dataset.__dict__,
                metrics=[],
                header_by_metric=header_by_metric,
                violin_values_by_sample_by_metric=dict(),
                scatter_values_by_sample_by_metric=dict(),
                all_samples=[],
                scatter_trace_params=dict(),
            )

            all_samples = set()
            for metric, header in header_by_metric.items():
                # Add Plotly-specific parameters to the header
                xaxis = {"ticksuffix": header.get("suffix")}
                header["xaxis"] = xaxis
                header["show_points"] = True
                header["show_only_outliers"] = False

                # Take non-empty values for the violin
                value_by_sample = values_by_sample_by_metric[metric]
                value_by_sample = {s: v for s, v in value_by_sample.items() if v is not None and str(v).strip() != ""}

                # Parse values to numbers if possible
                for s, v in value_by_sample.items():
                    if isinstance(v, str):
                        try:
                            value_by_sample[s] = int(v)
                        except ValueError:
                            try:
                                value_by_sample[s] = float(v)
                            except ValueError:
                                pass
                    if "modify" in header and callable(header["modify"]):
                        # noinspection PyBroadException
                        try:
                            value_by_sample[s] = header["modify"](v)
                        except Exception:  # User-provided modify function can raise any exception
                            pass

                if not value_by_sample:
                    logger.debug(f"No non-empty values found for metric: {header['title']}")
                    continue

                values_are_numeric = all(isinstance(v, (int, float)) for v in value_by_sample.values())
                values_are_integer = all(isinstance(v, int) for v in value_by_sample.values())
                if values_are_numeric:
                    # Remove NaN and Inf values
                    value_by_sample = {s: v for s, v in value_by_sample.items() if np.isfinite(v)}
                    if not value_by_sample:
                        logger.warning(f"All values are NaN or Inf for metric: {header['title']}")
                        continue

                header["show_points"] = len(value_by_sample) <= config.violin_min_threshold_no_points
                header["show_only_outliers"] = len(value_by_sample) > config.violin_min_threshold_outliers

                if values_are_numeric:
                    # Calculate range
                    xmin = header.get("dmin")
                    xmax = header.get("dmax")
                    tickvals = None
                    if xmin == xmax == 0:  # Plotly will modify the 0:0 range to -1:1, and we want to keep it 0-centered
                        xmax = 1
                        tickvals = [0]
                    xaxis["tickvals"] = tickvals
                    if xmin is not None and xmax is not None:
                        if header["show_points"]:
                            # add extra padding to avoid clipping the points at range limits
                            xmin -= (xmax - xmin) * 0.005
                            xmax += (xmax - xmin) * 0.005
                        xaxis["range"] = [xmin, xmax]

                if not header["show_points"]:  # Do not add any interactive points
                    scatter_values_by_sample = {}
                elif not header["show_only_outliers"]:
                    scatter_values_by_sample = {}  # will use the violin values
                else:
                    if not values_are_numeric:
                        logger.debug(
                            f"Violin for '{header['title']}': sample number is {len(value_by_sample)} > {header['show_only_outliers']}. "
                            f"As values are not numeric, will not add any interactive points."
                        )
                        scatter_values_by_sample = {}
                    else:
                        logger.debug(
                            f"Violin for '{header['title']}': sample number is {len(value_by_sample)} > {header['show_only_outliers']}. "
                            f"Will add interactive points only for the outlier values."
                        )
                        samples = list(value_by_sample.keys())
                        values = list(value_by_sample.values())
                        outlier_statuses = find_outliers(
                            values,
                            minval=header.get("dmin"),
                            maxval=header.get("dmax"),
                            metric=header["title"],
                        )
                        logger.debug(
                            f"Violin for '{header['title']}': found {np.count_nonzero(outlier_statuses)} outliers"
                        )
                        scatter_values_by_sample = {
                            samples[idx]: values[idx] for idx in range(len(samples)) if outlier_statuses[idx]
                        }

                ds.scatter_values_by_sample_by_metric[metric] = scatter_values_by_sample

                # Now sort and downsample values to keep max 2000 points for each metric
                violin_values_by_sample = value_by_sample
                max_violin_points = config.violin_downsample_after
                if max_violin_points is not None and len(violin_values_by_sample) > max_violin_points:
                    logger.debug(
                        f"Violin for '{header['title']}': sample number is {len(violin_values_by_sample)}. "
                        f"Will downsample to max {max_violin_points} points."
                    )
                    samples = list(violin_values_by_sample.keys())
                    values = list(violin_values_by_sample.values())
                    indices = np.argsort(values)
                    indices = indices[:: int(math.ceil(len(indices) / max_violin_points))]
                    violin_values_by_sample = {samples[idx]: values[idx] for idx in indices}

                ds.violin_values_by_sample_by_metric[metric] = violin_values_by_sample

                # Clean up the header
                if values_are_numeric and not values_are_integer and "tt_decimals" in header:
                    header["hoverformat"] = f".{header['tt_decimals']}f"
                    del header["tt_decimals"]
                if "modify" in header and callable(header["modify"]):
                    del header["modify"]  # To keep the data JSON-serializable

                all_samples.update(set(list(scatter_values_by_sample.keys())))
                all_samples.update(set(list(violin_values_by_sample.keys())))
                ds.metrics.append(metric)

            ds.all_samples = sorted(all_samples)

            ds.trace_params.update(
                orientation="h",
                box={"visible": True},
                meanline={"visible": True},
                fillcolor="grey",
                line={"width": 2, "color": "grey"},
                opacity=0.5,
                points=False,  # Don't show points, we'll add them manually
                # The hover information is useful, but the formatting is ugly and not
                # configurable as far as I can see. Also, it's not possible to disable it,
                # so setting it to "points" as we don't show points, so it's effectively
                # disabling it.
                hoveron="points",
            )

            # If all violins are grey, make the dots blue to make it more clear that it's interactive
            # if some violins are color-coded, make the dots black to make them less distracting
            marker_color = "black" if any(h.get("color") for h in ds.header_by_metric.values()) else "#0b79e6"
            ds.scatter_trace_params = {
                "mode": "markers",
                "marker": {
                    "size": 4,
                    "color": marker_color,
                },
                "showlegend": False,
                "hovertemplate": ds.trace_params["hovertemplate"],
                "hoverlabel": {"bgcolor": "white"},
            }
            return ds

        def create_figure(
            self,
            layout: go.Layout,
            is_log=False,
            is_pct=False,
            add_scatter=True,
        ):
            """
            Create a Plotly figure for a dataset
            """
            metrics = [m for m in self.metrics if not self.header_by_metric[m].get("hidden", False)]

            layout = copy.deepcopy(layout)
            layout.grid.rows = len(metrics)
            layout.grid.subplots = [[(f"x{i + 1}y{i + 1}" if i > 0 else "xy")] for i in range(len(metrics))]
            layout.height = ViolinPlot.VIOLIN_HEIGHT * len(metrics) + ViolinPlot.EXTRA_HEIGHT

            for metric_idx, metric in enumerate(metrics):
                header = self.header_by_metric[metric]

                layout[f"xaxis{metric_idx + 1}"] = {
                    "automargin": layout["xaxis"]["automargin"],
                    "color": layout["xaxis"]["color"],
                    "gridcolor": layout["xaxis"]["gridcolor"],
                    "zerolinecolor": layout["xaxis"]["zerolinecolor"],
                    "hoverformat": layout["xaxis"]["hoverformat"],
                    "tickfont": copy.deepcopy(layout["xaxis"]["tickfont"]),
                }
                layout[f"xaxis{metric_idx + 1}"].update(header.get("xaxis", {}))
                layout[f"yaxis{metric_idx + 1}"] = {
                    "automargin": layout["yaxis"]["automargin"],
                    "color": layout["yaxis"]["color"],
                    "gridcolor": layout["yaxis"]["gridcolor"],
                    "zerolinecolor": layout["yaxis"]["zerolinecolor"],
                    "hoverformat": layout["yaxis"]["hoverformat"],
                    "tickfont": copy.deepcopy(layout["yaxis"]["tickfont"]),
                }

                padding = "  "  # otherwise the labels will stick too close to the axis
                title = header["title"] + padding
                if header.get("namespace"):
                    title = f"{header['namespace']}{padding}<br>" + title
                layout[f"yaxis{metric_idx + 1}"].update(
                    {
                        "tickmode": "array",
                        "tickvals": [metric_idx],
                        "ticktext": [title],
                    }
                )

                if "hoverformat" in header:
                    layout[f"xaxis{metric_idx + 1}"]["hoverformat"] = header["hoverformat"]

                if header.get("color"):
                    layout[f"yaxis{metric_idx + 1}"]["tickfont"] = {
                        "color": f"rgb({header['color']})",
                    }

            layout["xaxis"] = layout["xaxis1"]
            layout["yaxis"] = layout["yaxis1"]

            fig = go.Figure(layout=layout)

            violin_values_by_sample_by_metric = self.violin_values_by_sample_by_metric

            for metric_idx, metric in enumerate(metrics):
                header = self.header_by_metric[metric]
                params = copy.deepcopy(self.trace_params)
                color = header.get("color")
                if color:
                    params["fillcolor"] = f"rgb({color})"
                    params["line"]["color"] = f"rgb({color})"

                violin_values_by_sample = violin_values_by_sample_by_metric[metric]
                axis_key = "" if metric_idx == 0 else str(metric_idx + 1)
                fig.add_trace(
                    go.Violin(
                        x=list(violin_values_by_sample.values()),
                        name=metric_idx,
                        text=list(violin_values_by_sample.keys()),
                        xaxis=f"x{axis_key}",
                        yaxis=f"y{axis_key}",
                        **params,
                    ),
                )

                if add_scatter and header["show_points"]:
                    if header["show_only_outliers"]:
                        scatter_values_by_sample = self.scatter_values_by_sample_by_metric[metric]
                    else:
                        scatter_values_by_sample = violin_values_by_sample
                    scatter_params = copy.deepcopy(self.scatter_trace_params)
                    for sample, value in scatter_values_by_sample.items():
                        # add vertical jitter (not working in python version currently)
                        y = float(metric_idx)
                        # y += random.uniform(-0.2, 0.2)
                        # y += random.random() * 0.3 - 0.3 / 2
                        fig.add_trace(
                            go.Scatter(
                                x=[value],
                                y=[y],
                                text=[sample],
                                xaxis=f"x{axis_key}",
                                yaxis=f"y{axis_key}",
                                **scatter_params,
                            ),
                        )
            return fig

    def __init__(
        self,
        list_of_values_by_sample_by_metric: List[Dict[str, Dict[str, Union[List[int], List[float], List[str]]]]],
        list_of_header_by_metric: List[Dict[str, Dict]],
        pconfig: Dict,
        dt: Optional[DataTable] = None,
        show_table_by_default: bool = False,
    ):
        assert len(list_of_values_by_sample_by_metric) == len(list_of_header_by_metric)
        assert len(list_of_values_by_sample_by_metric) > 0

        super().__init__(
            PlotType.VIOLIN,
            pconfig,
            n_datasets=len(list_of_values_by_sample_by_metric),
            id=dt.id if dt else None,
        )

        self.dt = dt
        self.no_violin = pconfig.get("no_violin", pconfig.get("no_beeswarm", False))
        self.show_table = dt is not None
        self.show_table_by_default = show_table_by_default or self.no_violin
        if dt:  # to make it different from the violin id
            dt.id = "table-" + dt.id

        self.datasets: List[ViolinPlot.Dataset] = [
            ViolinPlot.Dataset.create(ds, values_by_sample_by_metric, headers_by_metric)
            for ds, values_by_sample_by_metric, headers_by_metric in zip(
                self.datasets,
                list_of_values_by_sample_by_metric,
                list_of_header_by_metric,
            )
        ]

        # Violin-specific layout parameters
        self.layout.update(
            margin=dict(
                pad=0,
                b=40,
            ),
            xaxis=dict(
                # so Plotly doesn't try to fit the ticks the on the most bottom violin,
                # and squish the other violins
                automargin=False,
                tickfont=dict(size=9, color="rgba(0,0,0,0.5)"),
                gridcolor="rgba(0,0,0,0.1)",
                zerolinecolor="rgba(0,0,0,0.1)",
            ),
            yaxis=dict(
                tickfont=dict(size=9, color="rgba(0,0,0,0.5)"),
                gridcolor="rgba(0,0,0,0.1)",
                zerolinecolor="rgba(0,0,0,0.1)",
            ),
            violingap=0,
            grid=dict(
                columns=1,
                roworder="top to bottom",
                ygap=0.4,
            ),
            showlegend=False,
        )

        # If the number of samples is high:
        # - do not add a table
        # - plot a Violin in Python, and serialise the figure instead of the datasets
        self.n_samples = max(len(ds.all_samples) for ds in self.datasets)
        self.serialize_figure = False
        if self.n_samples > config.max_table_rows and not self.no_violin:
            self.show_table = False
            if self.show_table_by_default:
                logger.debug(
                    f"Table '{self.id}': sample number {self.n_samples} > {config.max_table_rows}, "
                    "Will render only a violin plot instead of the table"
                )

    @staticmethod
    def from_dt(dt: DataTable, show_table_by_default=False) -> "ViolinPlot":
        values_by_sample_by_metric = dict()
        header_by_metric = dict()

        for idx, k, header in dt.get_headers_in_order():
            rid = header["rid"]
            header_by_metric[rid] = {
                "namespace": header["namespace"],
                "title": header["title"],
                "description": header["description"],
                "dmax": header.get("dmax"),
                "dmin": header.get("dmin"),
                "suffix": header.get("suffix", ""),
                "color": header.get("colour", header.get("color")),
                "hidden": header.get("hidden"),
                "modify": header.get("modify"),
                "tt_decimals": header.get("decimalPlaces", 2),
            }
            values_by_sample_by_metric[rid] = dict()
            for s_name, val_by_metric in dt.data[idx].items():
                if k in val_by_metric:
                    values_by_sample_by_metric[rid][s_name] = val_by_metric[k]

        # If all colors are the same, remove them
        if len(set([v["color"] for v in header_by_metric.values()])) == 1:
            for v in header_by_metric.values():
                v.pop("color", None)

        # If all namespaces are the same as well, remove them too (usually they follow the colors pattern)
        if len(set([v["namespace"] for v in header_by_metric.values()])) == 1:
            for v in header_by_metric.values():
                v.pop("namespace", None)

        return ViolinPlot(
            [values_by_sample_by_metric],
            [header_by_metric],
            pconfig=dt.pconfig,
            dt=dt,
            show_table_by_default=show_table_by_default,
        )

    @staticmethod
    def tt_label() -> str:
        return ": %{x}"

    def add_to_report(self, report) -> str:
        warning = ""
        if self.show_table_by_default and not self.show_table:
            warning = (
                f'<p class="text-muted" id="table-violin-info-{self.id}">'
                + '<span class="glyphicon glyphicon-exclamation-sign" '
                + 'title="An interactive table is not available because of the large number of samples. '
                + "A violin plot is generated instead, showing density of values for each metric, as "
                + "well as hoverable points for outlier samples in each metric. "
                + 'See http://multiqc.info/docs/#tables--beeswarm-plots"'
                + f' data-toggle="tooltip"></span> Showing {self.n_samples} samples.</p>'
            )
        elif not self.show_table:
            warning = (
                f'<p class="text-muted" id="table-violin-info-{self.id}">'
                + '<span class="glyphicon glyphicon-exclamation-sign" '
                + 'title="An interactive table is not available because of the large number of samples. '
                + "The violin plot displays hoverable points only for outlier samples in each metric, "
                + "and the hiding/highlighting functionality through the toolbox only works for outliers "
                + 'See http://multiqc.info/docs/#tables--beeswarm-plots"'
                + f' data-toggle="tooltip"></span> Showing {self.n_samples} samples.</p>'
            )

        if not self.show_table:
            # Show violin alone.
            # Note that "no_violin" will be ignored here as we need to render _something_. The only case it can
            # happen if violin.plot() is called directly, and "no_violin" is passed, which doesn't make sense.
            html = warning + super().add_to_report(report)
        elif self.no_violin:
            # Show table alone
            table_html, configuration_modal = make_table(self.dt)
            html = warning + table_html + configuration_modal
        else:
            # Render both, add a switch between table and violin
            table_html, configuration_modal = make_table(self.dt, violin_id=self.id)
            violin_html = super().add_to_report(report)

            violin_visibility = "style='display: none;'" if self.show_table_by_default else ""
            html = f"<div id='mqc_violintable_wrapper_{self.id}' {violin_visibility}>{warning}{violin_html}</div>"

            table_visibility = "style='display: none;'" if not self.show_table_by_default else ""
            html += f"<div id='mqc_violintable_wrapper_{self.dt.id}' {table_visibility}>{table_html}</div>"

            html += configuration_modal

        return html

    def dump_for_javascript(self):
        """Serialise the data to pick up in Plotly-JS"""
        d = super().dump_for_javascript()
        d.update(
            {
                "static": self.show_table and self.show_table_by_default,
                "violin_height": ViolinPlot.VIOLIN_HEIGHT,
                "extra_height": ViolinPlot.EXTRA_HEIGHT,
            }
        )
        return d

    def buttons(self) -> []:
        """Add a control panel to the plot"""
        buttons = []
        if not self.flat and any(len(ds.metrics) > 1 for ds in self.datasets) and self.dt.id is not None:
            buttons.append(
                self._btn(
                    cls="mqc_table_configModal_btn",
                    label="<span class='glyphicon glyphicon-th'></span> Configure columns",
                    data_attrs={"toggle": "modal", "target": f"#{self.dt.id}_configModal"},
                )
            )
        if self.show_table:
            buttons.append(
                self._btn(
                    cls="mqc-violin-to-table",
                    label="<span class='glyphicon glyphicon-th-list'></span> Table",
                    data_attrs={"table-id": self.dt.id, "violin-id": self.id},
                )
            )

        return buttons + super().buttons()

    def save_data_file(self, dataset: Dataset) -> None:
        data = {}
        for metric in dataset.metrics:
            values_by_sample = dataset.violin_values_by_sample_by_metric[metric]
            title = dataset.header_by_metric[metric]["title"]
            for sample, value in values_by_sample.items():
                data.setdefault(sample, {})[title] = value

        util_functions.write_data_file(data, dataset.uid)


def find_outliers(
    values: Union[List[int], List[float]],
    top_n: Optional[int] = None,
    z_cutoff: float = 2.0,
    minval: Optional[Union[float, int]] = None,
    maxval: Optional[Union[float, int]] = None,
    metric: Optional[str] = None,
) -> np.array:
    """
    If `n` is defined, find `n` most outlying points in a list.
    Otherwise, find outliers with a Z-score above `z_cutoff`.

    Return a list of booleans, indicating if the value with this index is an outlier.

    `minval` and/or `maxval` can be added to include them into the outlier detection array.
    This is a trick to avoid the following problem: for example, percentage values are
    clustered around 0, but differ in e.g. 3+rd decimal place. In this case, the outliers
    will be nearly no different from the other values. If we add "100%" as an artificial
    additional value, none of those near-zero clustered values won't be called outliers.
    """
    if len(values) == 0 or (top_n is not None and top_n <= 0):
        return np.zeros(len(values), dtype=bool)

    added_values = []
    if minval is not None:
        added_values.append(minval)
    if maxval is not None:
        added_values.append(maxval)
    values = np.array(values + added_values)

    # Calculate the mean and standard deviation
    mean = np.mean(values)
    std_dev = np.std(values)
    if std_dev == 0:
        logger.debug(f"All {len(values)} points have the same values" + (f", metric: '{metric}'" if metric else ""))
        return np.zeros(len(values), dtype=bool)

    # Calculate Z-scores (measures of "outlyingness")
    z_scores = np.abs((values - mean) / std_dev)

    # Get indices of the top N outliers
    outlier_status = np.zeros(len(values), dtype=bool)
    if top_n:
        outlier_status[np.argsort(z_scores)[-top_n:]] = True
    else:
        indices = np.where(z_scores > z_cutoff)[0]
        while len(indices) <= len(added_values) and z_cutoff > 1.0:
            new_z_cutoff = z_cutoff - 0.2
            logger.debug(
                f"No outliers found with Z-score cutoff {z_cutoff:.1f}, trying a lower cutoff: {new_z_cutoff:.1f}"
                + (f", metric: '{metric}'" if metric else "")
            )
            z_cutoff = new_z_cutoff
            indices = np.where(z_scores > z_cutoff)[0]
        outlier_status[indices] = True

    if added_values:
        outlier_status = outlier_status[: -len(added_values)]
    return outlier_status



import logging

from multiqc.modules.base_module import BaseMultiqcModule

# Initialise the logger
log = logging.getLogger(__name__)


NAMESPACE = "Ploidy estimation"


class DragenPloidyEstimationMetrics(BaseMultiqcModule):
    def add_ploidy_estimation_metrics(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/ploidy_estimation_metrics"):
            data = parse_ploidy_estimation_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="ploidy_estimation_metrics")
            data_by_sample[s_name] = data

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)
        if not data_by_sample:
            return set()

        # Write data to file
        self.write_data_file(data_by_sample, "dragen_ploidy")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        headers = {
            "Ploidy estimation": {
                "title": "Sex",
                "description": "Sex chromosome ploidy estimation (XX, XY, X0, 00, etc.)",
                "scale": "Set3",
            }
        }
        self.general_stats_addcols(data_by_sample, headers, namespace=NAMESPACE)
        return data_by_sample.keys()


def parse_ploidy_estimation_metrics_file(f):
    """
    T_SRR7890936_50pc.ploidy_estimation_metrics.csv

    PLOIDY ESTIMATION,,Autosomal median coverage,55.63
    PLOIDY ESTIMATION,,X median coverage,27.44
    PLOIDY ESTIMATION,,Y median coverage,0.00
    PLOIDY ESTIMATION,,X median / Autosomal median,0.49
    PLOIDY ESTIMATION,,Y median / Autosomal median,0.00
    PLOIDY ESTIMATION,,Ploidy estimation,X0
    """

    data = {}
    for line in f["f"].splitlines():
        _, _, metric, stat = line.split(",")
        if stat.strip() == "":
            continue
        try:
            stat = float(stat)
        except ValueError:
            pass
        data[metric] = stat

    return data



"""
========
 simple
========

This theme attempts to generate a report which is a simple
as possible - no JavaScript where possible. The resulting
report is hopefully suitable for e-mailing, converting to PDF
and printing.

"""
import os

from multiqc.utils import config

template_parent = "default"

config.plots_force_flat = True
config.simple_output = True

template_dir = os.path.dirname(__file__)
base_fn = "base.html"



""" MultiQC submodule to parse output from RSeQC read_distribution.py
http://rseqc.sourceforge.net/#read-distribution-py """

import logging
import re

from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC read_distribution reports and parse their data"""

    # Set up vars
    self.read_dist = dict()
    first_regexes = {
        "total_reads": r"Total Reads\s+(\d+)\s*",
        "total_tags": r"Total Tags\s+(\d+)\s*",
        "total_assigned_tags": r"Total Assigned Tags\s+(\d+)\s*",
    }
    second_regexes = {
        "cds_exons": r"CDS_Exons\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "5_utr_exons": r"5'UTR_Exons\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "3_utr_exons": r"3'UTR_Exons\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "introns": r"Introns\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tss_up_1kb": r"TSS_up_1kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tss_up_5kb": r"TSS_up_5kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tss_up_10kb": r"TSS_up_10kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tes_down_1kb": r"TES_down_1kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tes_down_5kb": r"TES_down_5kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
        "tes_down_10kb": r"TES_down_10kb\s+(\d+)\s+(\d+)\s+([\d\.]+)\s*",
    }

    # Go through files and parse data using regexes
    for f in self.find_log_files("rseqc/read_distribution"):
        d = dict()
        for k, r in first_regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[k] = int(r_search.group(1))
        for k, r in second_regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[f"{k}_total_bases"] = int(r_search.group(1))
                d[f"{k}_tag_count"] = int(r_search.group(2))
                d[f"{k}_tags_kb"] = float(r_search.group(3))

        d["other_intergenic_tag_count"] = d["total_tags"] - d["total_assigned_tags"]

        # Calculate some percentages for parsed file
        if "total_tags" in d:
            t = float(d["total_tags"])
            pcts = dict()
            for k in d:
                if k.endswith("_tag_count"):
                    pk = f"{k[:-10]}_tag_pct"
                    try:
                        pcts[pk] = (float(d[k]) / t) * 100.0
                    except ZeroDivisionError:
                        pcts[pk] = 0
            d.update(pcts)

        if len(d) > 0:
            if f["s_name"] in self.read_dist:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="read_distribution")
            self.read_dist[f["s_name"]] = d

    # Filter to strip out ignored sample names
    self.read_dist = self.ignore_samples(self.read_dist)

    if len(self.read_dist) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Fix double counting of TSS_up and TES_down
    prefixes = ("tss_up", "tes_down")
    sizes = ("10kb", "5kb", "1kb")
    suffixes = ("total_bases", "tag_count")
    combinations = [
        (prefix, big, small, suffix)
        for prefix in prefixes
        for suffix in suffixes
        for big, small in zip(sizes, sizes[1:])
    ]

    for sample_name, sample in self.read_dist.items():
        for prefix, big, small, suffix in combinations:
            try:
                big_val = sample.pop(f"{prefix}_{big}_{suffix}")
                small_val = sample[f"{prefix}_{small}_{suffix}"]
                sample[f"{prefix}_{small}-{big}_{suffix}"] = big_val - small_val
            except KeyError as e:
                log.debug(f"Field {e} is missing from {sample_name}")

    # Write to file
    self.write_data_file(self.read_dist, "multiqc_rseqc_read_distribution")

    # Plot bar graph of groups
    keys = {
        "cds_exons_tag_count": {"name": "CDS_Exons"},
        "5_utr_exons_tag_count": {"name": "5'UTR_Exons"},
        "3_utr_exons_tag_count": {"name": "3'UTR_Exons"},
        "introns_tag_count": {"name": "Introns"},
        "tss_up_1kb_tag_count": {"name": "TSS_up_1kb"},
        "tss_up_1kb-5kb_tag_count": {"name": "TSS_up_1kb-5kb"},
        "tss_up_5kb-10kb_tag_count": {"name": "TSS_up_5kb-10kb"},
        "tes_down_1kb_tag_count": {"name": "TES_down_1kb"},
        "tes_down_1kb-5kb_tag_count": {"name": "TES_down_1kb-5kb"},
        "tes_down_5kb-10kb_tag_count": {"name": "TES_down_5kb-10kb"},
        "other_intergenic_tag_count": {"name": "Other_intergenic"},
    }

    # Config for the plot
    pconfig = {
        "id": "rseqc_read_distribution_plot",
        "title": "RSeQC: Read Distribution",
        "ylab": "# Tags",
        "cpswitch_counts_label": "Number of Tags",
        "cpswitch_c_active": False,
    }

    if max([d["total_tags"] for d in self.read_dist.values()]) > 0:
        self.add_section(
            name="Read Distribution",
            anchor="rseqc-read_distribution",
            description='<a href="http://rseqc.sourceforge.net/#read-distribution-py" target="_blank">Read Distribution</a>'
            " calculates how mapped reads are distributed over genome features.",
            plot=bargraph.plot(self.read_dist, keys, pconfig),
        )
    else:
        self.add_section(
            name="Read Distribution",
            anchor="rseqc-read_distribution",
            description='<a href="http://rseqc.sourceforge.net/#read-distribution-py" target="_blank">Read Distribution</a>'
            " calculates how mapped reads are distributed over genome features.",
            content='<div class="alert alert-info">All samples had zero tags.</div>',
        )

    # Return number of samples found
    return len(self.read_dist)



""" MultiQC module to parse output from pbmarkdup"""

import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    pbmarkdup module class, parses pbmarkdup output.
    """

    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="pbmarkdup",
            anchor="pbmarkdup",
            href="https://github.com/PacificBiosciences/pbmarkdup",
            info=(
                """
                 takes one or multiple sequencing chips of an amplified libray
                 as HiFi reads and marks or removes duplicates.
                 """
            ),
            # Can't find a DOI // doi=
        )

        self.pbmarkdup = dict()

        for logfile in self.find_log_files("pbmarkdup", filehandles=True):
            data = self.parse_logfile(logfile)
            if data:
                # Clean the file name
                s_name = self.clean_s_name(logfile["s_name"], logfile)
                self.pbmarkdup[s_name] = data
                self.add_data_source(logfile)

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names
        self.pbmarkdup = self.ignore_samples(self.pbmarkdup)

        # Raise ModuleNoSamplesFound if we did not find any data
        if not self.pbmarkdup:
            raise ModuleNoSamplesFound

        # Write the parsed data to file
        self.write_data_file(self.pbmarkdup, "multiqc_pbmarkdup")

        self.pbmarkdup_add_general_stats()

    def parse_logfile(self, logfile):
        """
        Parse the standard output from pbmarkdup

        This is currently quite ugly, since pbmarkdup does not have an easily
        parsable output format. See this github issue:
        https://github.com/PacificBiosciences/pbbioconda/issues/365

        For now, we just assume that the data fields are always in the same
        order, without any missing values.
        """

        file_content = logfile["f"]

        # The number of spaces in the header can vary, based on the length of
        # the name of the library. We therefore need a regex to match the
        # header.
        header_pattern = "LIBRARY +READS +UNIQUE MOLECULES +DUPLICATE READS"
        pattern = re.compile(header_pattern)

        # The file header
        file_header = next(file_content).strip()

        # Log an error if the header doesn't match the expected pattern
        if not re.match(pattern, file_header):
            fname = logfile["fn"]
            log.error(f"Can't parse file '{fname}', unknown header: '{file_header}'")
            return False

        data = dict()

        # Each parsable line is either for a library, or for 'TOTAL', the sum
        # of all parsed libraries
        for line in file_content:
            # Lines which start with --- denote separators in the file, and do
            # not need to be parsed
            if line.startswith("-"):
                continue

            # Not very nice, we assume that all fields are always present
            lib_name, reads, unique_mol_count, unique_mol_perc, duplicate_count, duplicate_perc = line.split()

            # We are only interested in the counts, not the percentages
            data[lib_name] = {
                "READS": int(reads),
                "UNIQUE MOLECULES": int(unique_mol_count),
                "DUPLICATE READS": int(duplicate_count),
            }

        return data

    def pbmarkdup_add_general_stats(self):
        """Add pbmarkdup duplicates to the general stats table"""

        general_stats_headers = {
            "unique_molecules": {
                "id": "unique_molecules",
                "title": "% Unique Molecules",
                "description": "Percentage of unique molecules",
                "suffix": "%",
                "min": 0,
                "max": 100,
                "modify": lambda x: x * 100,
                "scale": "RdYlGn",
            },
            "duplicate_reads": {
                "id": "duplicate_erads",
                "title": "% Duplicate Reads",
                "description": "Percentage of duplicate reads",
                "suffix": "%",
                "min": 0,
                "max": 100,
                "modify": lambda x: x * 100,
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
        }

        general = dict()

        for sample, sample_stats in self.pbmarkdup.items():
            # We are only interested in the total counts per sample
            total = sample_stats["TOTAL"]

            # Calculate the percentages unique and duplicated
            perc_unique = total["UNIQUE MOLECULES"] / total["READS"]
            perc_duplicate = total["DUPLICATE READS"] / total["READS"]

            general[sample] = {"unique_molecules": perc_unique, "duplicate_reads": perc_duplicate}

        self.general_stats_addcols(general, general_stats_headers)



""" MultiQC module to parse output from Conpair """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Conpair module class.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Conpair",
            anchor="conpair",
            href="https://github.com/nygenome/Conpair",
            info="is a fast and robust method dedicated for human tumor-normal "
            "studies to perform concordance verification, as well as "
            "cross-individual contamination level estimation in "
            "whole-genome and whole-exome sequencing experiments.",
            doi="bioinformatics/btw389",
        )

        self.conpair_data = dict()

        for f in self.find_log_files("conpair/concordance"):
            self.parse_conpair_logs(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        for f in self.find_log_files("conpair/contamination"):
            self.parse_conpair_logs(f)

        # Filter to strip out ignored sample names
        self.conpair_data = self.ignore_samples(self.conpair_data)

        if len(self.conpair_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.conpair_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.conpair_data, "multiqc_conpair")

        # Basic Stats Table
        self.conpair_general_stats_table()

    def parse_conpair_logs(self, f):
        """Go through log file looking for conpair concordance or contamination output
        One parser to rule them all."""

        conpair_regexes = {
            "concordance_concordance": [r"Concordance: ([\d\.]+)%\nBased.*", r"(\d\.\d+)\nBased.*"],
            "concordance_used_markers": [r"Based on (\d+)/\d+ markers"],
            "concordance_total_markers": [r"Based on \d+/(\d+) markers"],
            "concordance_marker_threshold": [r"\(coverage per marker threshold : (\d+) reads\)"],
            "concordance_min_mapping_quality": [r"Minimum mappinq quality: (\d+)"],
            "concordance_min_base_quality": [r"Minimum base quality: (\d+)"],
            "contamination_normal": [r"Normal sample contamination level: ([\d\.]+)%"],
            "contamination_tumor": [r"Tumor sample contamination level: ([\d\.]+)%"],
        }

        parsed_data = {}
        for k, r_arr in conpair_regexes.items():
            for r in r_arr:
                match = re.search(r, f["f"])
                if match:
                    parsed_data[k] = float(match.group(1))
                    if k == "concordance_concordance" and "Concordance" not in r:
                        parsed_data[k] = 100.0 * float(parsed_data[k])
                    break

        def _cp_type(data):
            if "concordance_concordance" in parsed_data:
                return "concordance"
            elif "contamination_normal" in parsed_data:
                return "contamination"

        if len(parsed_data) > 0:
            if f["s_name"] in self.conpair_data:
                if _cp_type(self.conpair_data[f["s_name"]]) == _cp_type(parsed_data):
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            else:
                self.conpair_data[f["s_name"]] = dict()
            self.add_data_source(f, section=_cp_type(parsed_data))
            self.conpair_data[f["s_name"]].update(parsed_data)

    def conpair_general_stats_table(self):
        """Take the parsed stats from the Conpair report and add it to the
        basic stats table at the top of the report"""

        headers = {}
        headers["concordance_concordance"] = {
            "title": "Concordance",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.2f}",
            "scale": "RdYlGn",
        }
        headers["contamination_normal"] = {
            "title": "N Contamination",
            "description": "Normal sample contamination level",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.3f}",
            "scale": "RdYlBu-rev",
        }
        headers["contamination_tumor"] = {
            "title": "T Contamination",
            "description": "Tumor sample contamination level",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.3f}",
            "scale": "RdYlBu-rev",
        }
        self.general_stats_addcols(self.conpair_data, headers)



#!/usr/bin/env python

""" MultiQC functions to plot a heatmap """


import logging
import random

from multiqc.utils import config, report

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, xcats, ycats, pconfig=None):
    """Plot a 2D heatmap.
    :param data: List of lists, each a representing a row of values.
    :param xcats: Labels for x axis
    :param ycats: Labels for y axis. Defaults to same as x.
    :param pconfig: optional dict with config key:value pairs.
    :return: HTML and JS, ready to be inserted into the page
    """
    return highcharts_heatmap(data, xcats, ycats, pconfig)


def highcharts_heatmap(data, xcats, ycats, pconfig=None):
    """
    Build the HTML needed for a HighCharts line graph. Should be
    called by plot_xy_data, which properly formats input data.
    """
    if pconfig is None:
        pconfig = {}

    # Reformat the data for highcharts
    pdata = []
    minval = None
    maxval = None
    for i, arr in enumerate(data):
        for j, val in enumerate(arr):
            pdata.append([j, i, val])
            if isinstance(val, (int, float)):
                if minval is None or val < minval:
                    minval = val
                if maxval is None or val > maxval:
                    maxval = val

    if "min" not in pconfig:
        pconfig["min"] = minval
    if "max" not in pconfig:
        pconfig["max"] = maxval

    # Get the plot ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_hcplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Build the HTML for the page
    html = """
    <div class="mqc_hcplot_plotgroup">
        <div class="mqc_hcplot_range_sliders">
            <div>
                <label for="{id}_range_slider_min_txt">Min:</label>
                <input id="{id}_range_slider_min_txt" type="number" class="form-control" value="{min}" min="{min}" max="{max}" data-minmax="min" data-target="{id}" />
                <input id="{id}_range_slider_min" type="range" value="{min}" min="{min}" max="{max}" step="any" data-minmax="min" data-target="{id}" />
            </div>
            <div>
                <label for="{id}_range_slider_max_txt">Max:</label>
                <input id="{id}_range_slider_max_txt" type="number" class="form-control" value="{max}" min="{min}" max="{max}" data-minmax="max" data-target="{id}" />
                <input id="{id}_range_slider_max" type="range" value="{max}" min="{min}" max="{max}" step="any" data-minmax="max" data-target="{id}" />
            </div>
        </div>
        <div class="hc-plot-wrapper"{height}>
            <div id="{id}" class="hc-plot not_rendered hc-heatmap">
                <small>loading..</small>
            </div>
        </div>
    </div> \n""".format(
        id=pconfig["id"],
        min=pconfig["min"],
        max=pconfig["max"],
        height=f' style="height:{pconfig["height"]}px"' if "height" in pconfig else "",
    )

    report.num_hc_plots += 1

    report.plot_data[pconfig["id"]] = {
        "plot_type": "heatmap",
        "data": pdata,
        "xcats": xcats,
        "ycats": ycats,
        "config": pconfig,
    }

    return html



# -*- coding: utf-8 -*-

""" MultiQC submodule to parse output from GATK BaseRecalibrator """

import logging
from collections import namedtuple
from itertools import groupby

from multiqc.plots import linegraph, scatter

# Initialise the logger
log = logging.getLogger(__name__)

RecalTableType = namedtuple("RecalTableType", "pre_recalibration post_recalibration".split())
recal_table_type = RecalTableType(0, 1)


class BaseRecalibratorMixin:
    def parse_gatk_base_recalibrator(self):
        """Find GATK BaseRecalibrator logs and parse their data"""

        report_table_headers = {
            "#:GATKTable:Arguments:Recalibration argument collection values used in this run": "arguments",
            "#:GATKTable:Quantized:Quality quantization map": "quality_quantization_map",
            "#:GATKTable:RecalTable0:": "recal_table_0",
            "#:GATKTable:RecalTable1:": "recal_table_1",
            "#:GATKTable:RecalTable2:": "recal_table_2",
        }
        samples_kept = {rt_type: set() for rt_type in recal_table_type}
        self.gatk_base_recalibrator = {
            recal_type: {table_name: {} for table_name in report_table_headers.values()}
            for recal_type in recal_table_type
        }

        for f in self.find_log_files("gatk/base_recalibrator", filehandles=True):
            # Check that we're not ignoring this sample name
            if self.is_ignore_sample(f["s_name"]):
                continue

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

            parsed_data = self.parse_report(f["f"].readlines(), report_table_headers)
            rt_type = determine_recal_table_type(parsed_data)
            if len(parsed_data) > 0:
                if f["s_name"] in samples_kept[rt_type]:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                samples_kept[rt_type].add(f["s_name"])

                self.add_data_source(f, section="base_recalibrator")
                for table_name, sample_tables in parsed_data.items():
                    self.gatk_base_recalibrator[rt_type][table_name][f["s_name"]] = sample_tables

        # Filter to strip out ignored sample names. Again.
        for rt_type in recal_table_type:
            for table_name, sample_tables in self.gatk_base_recalibrator[rt_type].items():
                self.gatk_base_recalibrator[rt_type][table_name] = self.ignore_samples(sample_tables)

        n_reports_found = sum([len(samples_kept[rt_type]) for rt_type in recal_table_type])

        if n_reports_found > 0:
            log.info(f"Found {n_reports_found} BaseRecalibrator reports")

            # Write data to file
            self.write_data_file(self.gatk_base_recalibrator, "gatk_base_recalibrator")

            self.add_quality_score_vs_no_of_observations_section()
            self.add_reported_vs_empirical_section()

        return n_reports_found

    def add_quality_score_vs_no_of_observations_section(self):
        """Add a section for the quality score vs number of observations line plot"""

        sample_data = []
        data_labels = []

        # Loop through the different data types
        for rt_type_name, rt_type in recal_table_type._asdict().items():
            sample_tables = self.gatk_base_recalibrator[rt_type]["quality_quantization_map"]
            if len(sample_tables) == 0:
                continue

            count_data = {}
            pct_data = {}
            for sample, table in sample_tables.items():
                count_data[sample] = {}
                pct_data[sample] = {}

                # Get the total count for this sample
                sample_y_sum = sum(int(y) for y in table["Count"])

                # Collect the data for the plots
                for x, y in zip(table["QualityScore"], table["Count"]):
                    # Quality score counts
                    count_data[sample][int(x)] = int(y)
                    # Quality score percentages
                    try:
                        pct_data[sample][int(x)] = float(y) / sample_y_sum
                    except ZeroDivisionError:
                        pct_data[sample][int(x)] = 0

            # Append the datasets for this data type
            sample_data.append(count_data)
            sample_data.append(pct_data)

            # Build data label configs for this data type
            data_labels.append({"name": f"{rt_type_name.capitalize().replace('_', '-')} Count", "ylab": "Count"})
            data_labels.append({"name": f"{rt_type_name.capitalize().replace('_', '-')} Percent", "ylab": "Percent"})

        plot = linegraph.plot(
            sample_data,
            pconfig={
                "title": "GATK: Observed Quality Score Counts",
                "id": "gatk-base-recalibrator-quality-scores-plot",
                "xlab": "Observed Quality Score",
                "ylab": "Count",
                "xDecimals": False,
                "data_labels": data_labels,
            },
        )

        # Reported vs empirical quality scores
        self.add_section(
            name="Observed Quality Scores",
            anchor="gatk-base-recalibrator-quality-scores",
            description=(
                "This plot shows the distribution of base quality scores in each sample before and "
                "after base quality score recalibration (BQSR). Applying BQSR should broaden the "
                "distribution of base quality scores."
            ),
            helptext=(
                "For more information see "
                "[the Broad's description of BQSR]"
                "(https://gatkforums.broadinstitute.org/gatk/discussion/44/base-quality-score-recalibration-bqsr)"
                "."
            ),
            plot=plot,
        )

    def add_reported_vs_empirical_section(self):
        sample_data = []
        data_labels = []

        # Loop through the different data types
        for (
            rt_type_name,
            rt_type,
        ) in recal_table_type._asdict().items():
            # This table appears to be the correct one to use for reported vs empirical
            # https://github.com/broadinstitute/gatk/blob/853b53ec2a3ac2d90d7d82a6c8451e29a34692d2/src/main/resources/org/broadinstitute/hellbender/utils/recalibration/BQSR.R#L148
            sample_tables = self.gatk_base_recalibrator[rt_type]["recal_table_1"]
            if len(sample_tables) == 0:
                continue

            reported_empirical = {}
            for sample, table in sample_tables.items():
                reported_empirical[sample] = []
                table_rows = [dict(zip(table, r)) for r in zip(*table.values())]
                table_rows.sort(key=lambda r: r["QualityScore"])
                for reported, group in groupby(table_rows, lambda r: r["QualityScore"]):
                    g = list(group)
                    reported_empirical[sample].append(
                        {
                            "x": int(reported),
                            "y": sum(float(r["EmpiricalQuality"]) for r in g) / len(g) if len(g) > 0 else 0,
                        }
                    )

            sample_data.append(reported_empirical)

            # Build data label configs for this data type
            data_labels.append(
                {"name": f"{rt_type_name} Reported vs. Empirical Quality", "ylab": "Empirical quality score"}
            )

        plot = scatter.plot(
            sample_data,
            pconfig={
                "title": "Reported vs. Empirical Quality",
                "id": "gatk-base-recalibrator-reported-empirical-plot",
                "xlab": "Reported quality score",
                "ylab": "Empirical quality score",
                "xDecimals": False,
                "data_labels": data_labels,
                "xmin": 0,
                "ymin": 0,
                "square": True,
            },
        )

        self.add_section(
            name="Reported Quality vs. Empirical Quality",
            anchor="gatk-base-recalibrator-reported-empirical",
            description="Plot shows the reported quality score vs the empirical quality score.",
            plot=plot,
        )


def determine_recal_table_type(parsed_data):
    arguments = dict(zip(parsed_data["arguments"]["Argument"], parsed_data["arguments"]["Value"]))
    if arguments["recalibration_report"] == "null":
        return recal_table_type.pre_recalibration
    else:
        return recal_table_type.post_recalibration



""" MultiQC submodule to parse output from Picard MeanQualityByCycle"""

import logging

from multiqc.plots import linegraph

from .util import read_histogram

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find Picard QualityByCycleMetrics reports and parse their data"""

    headers = ["CYCLE", "MEAN_QUALITY"]
    formats = [int, float]
    all_data = read_histogram(
        self,
        "picard/quality_by_cycle",
        headers,
        formats,
        picard_tool="MeanQualityByCycle",
        sentieon_algo="MeanQualityByCycle",
    )

    if not all_data:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write parsed data to a file
    self.write_data_file(all_data, f"multiqc_{self.anchor}_quality_by_cycle")

    # Plot the data and add section
    pconfig = {
        "id": f"{self.anchor}_quality_by_cycle",
        "title": f"{self.name}: Mean Base Quality by Cycle",
        "ylab": "Mean Base Quality",
        "xlab": "Cycle Number",
        "xDecimals": False,
        "tt_label": "<b>cycle {point.x}</b>: {point.y:.2f}",
        "ymin": 0,
    }

    lg = {}
    for s_name in all_data:
        lg[s_name] = dict((cycle, data["MEAN_QUALITY"]) for cycle, data in all_data[s_name].items())

    self.add_section(
        name="Mean Base Quality by Cycle",
        anchor=f"{self.anchor}-quality-by-cycle",
        description="Plot shows the mean base quality by cycle.",
        helptext="""
        This metric gives an overall snapshot of sequencing machine performance.
        For most types of sequencing data, the output is expected to show a slight
        reduction in overall base quality scores towards the end of each read.

        Spikes in quality within reads are not expected and may indicate that technical
        problems occurred during sequencing.
        """,
        plot=linegraph.plot([lg], pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(all_data)



""" MultiQC module to parse output from Cutadapt """


import logging
import os
import re
import shlex
import json

from packaging import version

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Cutadapt module class, parses stderr logs.
    Also understands logs saved by Trim Galore!
    (which contain cutadapt logs)
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Cutadapt",
            anchor="cutadapt",
            href="https://cutadapt.readthedocs.io/",
            info="""is a tool to find and remove adapter sequences, primers, poly-A
                    tails and other types of unwanted sequence from your high-throughput
                    sequencing reads.""",
            doi="10.14806/ej.17.1.200",
        )

        # Find and load any Cutadapt reports
        self.cutadapt_data = dict()
        self.cutadapt_length_counts = {"default": dict()}
        self.cutadapt_length_exp = {"default": dict()}
        self.cutadapt_length_obsexp = {"default": dict()}

        for f in self.find_log_files("cutadapt"):
            self.parse_file(f)

        # Transform trimmed length data by type
        self.transform_trimming_length_data_for_plot()

        # Filter to strip out ignored sample names
        self.cutadapt_data = self.ignore_samples(self.cutadapt_data)

        if len(self.cutadapt_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.cutadapt_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.cutadapt_data, "multiqc_cutadapt")

        # Basic Stats Table
        self.cutadapt_general_stats_table()

        # Bar plot with number of reads trimmed
        self.cutadapt_filtered_barplot()

        # Trimming Length Profiles
        self.cutadapt_length_trimmed_plot()

    def parse_file(self, f):
        if f["fn"].endswith(".json"):
            self.parse_json(f)
        else:
            self.parse_log(f)

        # Calculate a few extra numbers of our own
        for s_name, d in self.cutadapt_data.items():
            # Percent trimmed
            if d.get("bp_processed"):
                if d.get("bp_written") is not None:
                    self.cutadapt_data[s_name]["percent_trimmed"] = (
                        float(d["bp_processed"] - d["bp_written"]) / d["bp_processed"]
                    ) * 100
                elif d.get("bp_trimmed") is not None:
                    self.cutadapt_data[s_name]["percent_trimmed"] = (
                        (float(d.get("bp_trimmed", 0)) + float(d.get("quality_trimmed", 0))) / d["bp_processed"]
                    ) * 100
            # Add missing filtering categories for pre-1.7 logs
            if version.parse(d["cutadapt_version"]) > version.parse("1.6"):
                if d.get("r_processed") is not None:
                    r_filtered_unexplained = (
                        d["r_processed"]
                        - (d.get("r_too_short") or 0)
                        - (d.get("r_too_long") or 0)
                        - (d.get("r_too_many_N") or 0)
                        - (d.get("r_written") or 0)
                    )
                    if r_filtered_unexplained > 0:
                        self.cutadapt_data[s_name]["r_filtered_unexplained"] = r_filtered_unexplained
                if d.get("pairs_processed") is not None:
                    pairs_filtered_unexplained = (
                        d["pairs_processed"]
                        - (d.get("pairs_too_short") or 0)
                        - (d.get("pairs_too_long") or 0)
                        - (d.get("pairs_too_many_N") or 0)
                        - (d.get("pairs_written") or 0)
                    )
                    if pairs_filtered_unexplained > 0:
                        self.cutadapt_data[s_name]["pairs_filtered_unexplained"] = pairs_filtered_unexplained

    def parse_json(self, f):
        path = os.path.join(f["root"], f["fn"])
        with open(path, "r") as fh:
            data = json.load(fh)

        s_name = self.clean_s_name([v for k, v in data["input"].items() if k.startswith("path")], f)
        if s_name in self.cutadapt_data:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")

        self.add_software_version(data["cutadapt_version"], s_name)
        self.add_data_source(f, s_name)

        d = dict()
        d["cutadapt_version"] = data["cutadapt_version"]
        d["bp_processed"] = data["basepair_counts"]["input"]
        d["bp_written"] = data["basepair_counts"]["output"]
        d["quality_trimmed"] = data["basepair_counts"]["quality_trimmed"]
        d["r_processed"] = data["read_counts"]["input"]
        d["r1_with_adapters"] = data["read_counts"]["read1_with_adapter"]
        d["r2_with_adapters"] = data["read_counts"]["read2_with_adapter"]
        d["r_too_short"] = data["read_counts"]["filtered"]["too_short"]
        d["r_too_long"] = data["read_counts"]["filtered"]["too_long"]
        d["r_too_many_N"] = data["read_counts"]["filtered"]["too_many_n"]
        d["r_written"] = data["read_counts"]["output"]
        d = {k: v for k, v in d.items() if v is not None}
        self.cutadapt_data[s_name] = d

        for end, end_key in [("5", "five_prime_end"), ("3", "three_prime_end")]:
            if end not in self.cutadapt_length_counts:
                self.cutadapt_length_counts[end] = dict()
                self.cutadapt_length_exp[end] = dict()
                self.cutadapt_length_obsexp[end] = dict()
            if s_name not in self.cutadapt_length_counts[end]:
                self.cutadapt_length_counts[end][s_name] = dict()
                self.cutadapt_length_exp[end][s_name] = dict()
                self.cutadapt_length_obsexp[end][s_name] = dict()

            for read in ["read1", "read2"]:
                if f"input_{read}" not in data["basepair_counts"]:
                    continue
                for adapter_data in data[f"adapters_{read}"]:
                    end_data = adapter_data.get(end_key)
                    if end_data:
                        for trimmed_length in end_data["trimmed_lengths"]:
                            length = trimmed_length["len"]
                            if length not in self.cutadapt_length_counts[end][s_name]:
                                self.cutadapt_length_counts[end][s_name][length] = 0
                                self.cutadapt_length_exp[end][s_name][length] = 0
                                self.cutadapt_length_obsexp[end][s_name][length] = 0
                            self.cutadapt_length_counts[end][s_name][length] += trimmed_length["counts"][0]
                            self.cutadapt_length_exp[end][s_name][length] += trimmed_length["expect"]
                            if trimmed_length["expect"] > 0:
                                self.cutadapt_length_obsexp[end][s_name][length] += (
                                    trimmed_length["counts"][0] / trimmed_length["expect"]
                                )

    def parse_log(self, f):
        """Go through log file looking for cutadapt output"""
        regexes = {
            "1.7": {
                "bp_processed": r"Total basepairs processed:\s*([\d,]+) bp",
                "bp_written": r"Total written \(filtered\):\s*([\d,]+) bp",
                "quality_trimmed": r"Quality-trimmed:\s*([\d,]+) bp",
                "r_processed": r"Total reads processed:\s*([\d,]+)",
                "pairs_processed": r"Total read pairs processed:\s*([\d,]+)",
                "r_with_adapters": r"Reads with adapters:\s*([\d,]+)",
                "r1_with_adapters": r"Read 1 with adapter:\s*([\d,]+)",
                "r2_with_adapters": r"Read 2 with adapter:\s*([\d,]+)",
                "r_too_short": r"Reads that were too short:\s*([\d,]+)",
                "pairs_too_short": r"Pairs that were too short:\s*([\d,]+)",
                "r_too_long": r"Reads that were too long:\s*([\d,]+)",
                "pairs_too_long": r"Pairs that were too long:\s*([\d,]+)",
                "r_too_many_N": r"Reads with too many N:\s*([\d,]+)",
                "pairs_too_many_N": r"Pairs with too many N:\s*([\d,]+)",
                "r_written": r"Reads written \(passing filters\):\s*([\d,]+)",
                "pairs_written": r"Pairs written \(passing filters\):\s*([\d,]+)",
            },
            "1.6": {
                "r_processed": r"Processed reads:\s*([\d,]+)",
                "bp_processed": r"Processed bases:\s*([\d,]+) bp",
                "r_trimmed": r"Trimmed reads:\s*([\d,]+)",
                "quality_trimmed": r"Quality-trimmed:\s*([\d,]+) bp",
                "bp_trimmed": r"Trimmed bases:\s*([\d,]+) bp",
                "too_short": r"Too short reads:\s*([\d,]+)",
                "too_long": r"Too long reads:\s*([\d,]+)",
            },
        }
        s_name = None
        end = "default"
        cutadapt_version = None
        parsing_version = "1.7"
        log_section = None
        path = os.path.join(f["root"], f["fn"])
        with open(path, "r", encoding="utf-8", errors="ignore") as fh:
            lines = iter(fh.readlines())
        for line in lines:
            # New log starting
            if "This is cutadapt" in line or "cutadapt version" in line:
                s_name = None
                end = "default"
                cutadapt_version = None
                c_version = re.match(r"This is cutadapt ([\d\.]+)", line)
                if c_version:
                    cutadapt_version = c_version.group(1)
                    try:
                        assert version.parse(c_version.group(1)) <= version.parse("1.6")
                        parsing_version = "1.6"
                    except Exception:
                        parsing_version = "1.7"
                c_version_old = re.match(r"cutadapt version ([\d\.]+)", line)
                if c_version_old:
                    cutadapt_version = c_version_old.group(1)
                    # The pattern "cutadapt version XX" is only pre-1.6
                    parsing_version = "1.6"
            # Get sample name from end of command line params
            cl_pref = "Command line parameters: "
            if line.startswith(cl_pref):
                input_fqs = []
                args = shlex.split(line[len(cl_pref) :])
                for i, x in enumerate(args):
                    if (
                        not x.startswith("-")
                        and x.endswith((".fastq", ".fq", ".gz", ".dat"))
                        and (i == 0 or args[i - 1] not in ["-o", "-p", "--output", "--paired-output"])
                    ):
                        input_fqs.append(x)
                if input_fqs:
                    s_name = self.clean_s_name(input_fqs, f)
                else:
                    # Manage case where sample name is '-' (reading from stdin)
                    s_name = f["s_name"]

                if s_name in self.cutadapt_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.cutadapt_data[s_name] = dict()
                if cutadapt_version:
                    self.cutadapt_data[s_name]["cutadapt_version"] = cutadapt_version

            if s_name is not None:
                # Add version info to module
                if cutadapt_version is not None:
                    self.add_software_version(cutadapt_version, s_name)

                self.add_data_source(f, s_name)

                # Search regexes for overview stats
                for k, r in regexes[parsing_version].items():
                    match = re.search(r, line)
                    if match:
                        self.cutadapt_data[s_name][k] = int(match.group(1).replace(",", ""))

                # Starting a new section
                if "===" in line:
                    log_section = line.strip().strip("=").strip()

                # Detect whether 3' or 5'
                end_regex = re.search(r"Type: regular (\d)'", line)
                if end_regex:
                    end = end_regex.group(1)

                if "Overview of removed sequences" in line:
                    if "' end" in line:
                        res = re.search(r"(\d)' end", line)
                        end = res.group(1)

                    # Initialise dictionaries for length data if not already done
                    if end not in self.cutadapt_length_counts:
                        self.cutadapt_length_counts[end] = dict()
                        self.cutadapt_length_exp[end] = dict()
                        self.cutadapt_length_obsexp[end] = dict()

                # Histogram showing lengths trimmed
                if "length" in line and "count" in line and "expect" in line:
                    plot_sname = s_name
                    if log_section is not None:
                        plot_sname = f"{s_name} - {log_section}"
                    self.cutadapt_length_counts[end][plot_sname] = dict()
                    self.cutadapt_length_exp[end][plot_sname] = dict()
                    self.cutadapt_length_obsexp[end][plot_sname] = dict()

                    # Nested loop to read this section while the regex matches
                    for line2 in lines:
                        r_seqs = re.search(r"^(\d+)\s+(\d+)\s+([\d\.]+)", line2)
                        if r_seqs:
                            a_len = int(r_seqs.group(1))
                            self.cutadapt_length_counts[end][plot_sname][a_len] = int(r_seqs.group(2))
                            self.cutadapt_length_exp[end][plot_sname][a_len] = float(r_seqs.group(3))
                            if float(r_seqs.group(3)) > 0:
                                self.cutadapt_length_obsexp[end][plot_sname][a_len] = float(r_seqs.group(2)) / float(
                                    r_seqs.group(3)
                                )
                            else:
                                # Cheating, I know. Infinity is difficult to plot.
                                self.cutadapt_length_obsexp[end][plot_sname][a_len] = float(r_seqs.group(2))
                        else:
                            break

    def transform_trimming_length_data_for_plot(self):
        """Check if we parsed double ended data and transform it accordingly"""
        # Sanity check
        if (
            self.cutadapt_length_counts.keys() != self.cutadapt_length_exp.keys()
            or self.cutadapt_length_exp.keys() != self.cutadapt_length_obsexp.keys()
        ):
            log.error("Something went wrong...")
            log.debug("Keys in trimmed length data differed")
            raise ModuleNoSamplesFound

        if len(self.cutadapt_length_counts["default"]) == 0:
            self.cutadapt_length_counts.pop("default")
            self.cutadapt_length_exp.pop("default")
            self.cutadapt_length_obsexp.pop("default")

    def cutadapt_general_stats_table(self):
        """Take the parsed stats from the Cutadapt report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "percent_trimmed": {
                "title": "% BP Trimmed",
                "description": "% Total Base Pairs trimmed",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlBu-rev",
            }
        }
        self.general_stats_addcols(self.cutadapt_data, headers)

    def cutadapt_filtered_barplot(self):
        """Bar plot showing proportion of reads trimmed"""

        pconfig = {"id": "cutadapt_filtered_reads_plot", "title": "Cutadapt: Filtered Reads", "ylab": "Counts"}

        # We just use all categories. If a report is generated with a mixture
        # of SE and PE data then this means quite a lot of categories.
        # Usually, only a single data type is used though - in that case
        # any categories with 0 across all samples will be ignored.
        cats = {
            "pairs_written": {"name": "Pairs passing filters"},
            "r_written": {"name": "Reads passing filters"},
            "pairs_too_short": {"name": "Pairs that were too short"},
            "r_too_short": {"name": "Reads that were too short"},
            "pairs_too_long": {"name": "Pairs that were too long"},
            "r_too_long": {"name": "Reads that were too long"},
            "pairs_too_many_N": {"name": "Pairs with too many N"},
            "r_too_many_N": {"name": "Reads with too many N"},
            "pairs_filtered_unexplained": {"name": "Filtered pairs (uncategorised)"},
            "r_filtered_unexplained": {"name": "Filtered reads (uncategorised)"},
        }

        self.add_section(
            name="Filtered Reads",
            anchor="cutadapt_filtered_reads",
            description="This plot shows the number of reads (SE) / pairs (PE) removed by Cutadapt.",
            plot=bargraph.plot(self.cutadapt_data, cats, pconfig),
        )

    def cutadapt_length_trimmed_plot(self):
        """Generate the trimming length plot"""
        ends = [end for end, d_by_s in self.cutadapt_length_counts.items() if any(d_by_s.values())]
        ends = [x for x in ["default", "5", "3"] if x in ends]  # enforcing order
        for end in ends:
            pconfig = {
                "id": f"cutadapt_trimmed_sequences_plot_{end}",
                "title": "Cutadapt: Lengths of Trimmed Sequences{}".format(
                    "" if end == "default" else f" ({end}' end)"
                ),
                "ylab": "Counts",
                "xlab": "Length Trimmed (bp)",
                "xDecimals": False,
                "ymin": 0,
                "tt_label": "<b>{point.x} bp trimmed</b>: {point.y:.0f}",
                "data_labels": [
                    {"name": "Counts", "ylab": "Count"},
                    {"name": "Obs/Exp", "ylab": "Observed / Expected"},
                ],
            }

            self.add_section(
                name="Trimmed Sequence Lengths{}".format("" if end == "default" else f" ({end}')"),
                anchor=f"cutadapt_trimmed_sequences{'' if end == 'default' else f'_{end}'}",
                description="This plot shows the number of reads with certain lengths of adapter trimmed{}.".format(
                    "" if end == "default" else f" for the {end}' end"
                ),
                helptext="""
                Obs/Exp shows the raw counts divided by the number expected due to sequencing errors.
                A defined peak may be related to adapter length.

                See the [cutadapt documentation](http://cutadapt.readthedocs.org/en/latest/guide.html#how-to-read-the-report)
                for more information on how these numbers are generated.
                """,
                plot=linegraph.plot([self.cutadapt_length_counts[end], self.cutadapt_length_obsexp[end]], pconfig),
            )



""" MultiQC module to parse output from fgbio GroupReadsByUmi
"""


import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class GroupReadsByUmiMixin:
    def parse_groupreadsbyumi(self):
        # Parse data
        self.parse_groupreadsbyumi_log()

        # Make a plot if we have anything
        if len(self.fgbio_umi_data) > 0:
            self.parse_groupreadsbyumi_plot()

        # Return the number of samples parsed
        return len(self.fgbio_umi_data)

    def parse_groupreadsbyumi_log(self):
        umi_data = dict()
        umi_data_normed = dict()

        for f in self.find_log_files("fgbio/groupreadsbyumi"):
            # add file to data sources
            self.add_data_source(f)
            family_size = []
            for line in f["f"].splitlines():
                if not line.startswith("family_size"):
                    family_size.append(tuple(line.split("\t")))

            umi_data[f["s_name"]] = {int(s): int(d[1]) for s, d in enumerate(family_size, 1)}
            umi_data_normed[f["s_name"]] = {int(s): float(d[2]) * 100.0 for s, d in enumerate(family_size, 1)}

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        # Filter samples
        self.fgbio_umi_data = self.ignore_samples(umi_data)
        self.fgbio_umi_data_normed = self.ignore_samples(umi_data_normed)

        # Write data to file
        self.write_data_file(self.fgbio_umi_data, "fgbio_umi")
        self.write_data_file(self.fgbio_umi_data_normed, "fgbio_umi_normed")

    def parse_groupreadsbyumi_plot(self):
        config = {
            "id": "fgbio-groupreadsbyumi-plot",
            "title": "fgbio: Family size count",
            "ylab": "Number of UMIs",
            "xlab": "Reads supporting UMI",
            "xmax": 15,
            "xDecimals": False,
            "data_labels": [
                {"name": "Counts", "ylab": "Number of UMIs"},
                {"name": "Percentages", "ylab": "Percentage of sample"},
            ],
        }

        self.add_section(
            name="GroupReadsByUmi statistics",
            anchor="fgbio-groupreadsbyumi",
            description="""During `GroupReadsByUmi` processing, family size count data is generated,
                             showing number of UMIs represented by a certain number of reads.""",
            helptext="""
            This tool groups reads together that appear to have come from the same original molecule.
            Reads are grouped by template, and then templates are sorted by the 5' mapping positions
            of the reads from the template, used from earliest mapping position to latest.
            Reads that have the same end positions are then sub-grouped by UMI sequence.

            The histogram shows tag family size counts or percentages.
            """,
            plot=linegraph.plot([self.fgbio_umi_data, self.fgbio_umi_data_normed], config),
        )



import logging
import re

from multiqc.plots import bargraph

log = logging.getLogger(__name__)

VERSION_REGEX = r"sambamba ([\d\.]+)"


class SambambaMarkdupMixin:

    """Find and parse Sambamba Markdup output log files and calculate duplication rate"""

    def parse_sambamba_markdup(self):
        # Clean sample name from 'markdup_sample_1' to 'sample_1'
        # Find and load sambamba logs to markdup_data.
        # Regex for key phrases and calculate duplicate rate.
        # Detect if sample is single or paired-end reads. Process differently.
        # Give user warning if redundant samples are found.

        self.markdup_data = dict()

        for f in self.find_log_files("sambamba/markdup"):
            if f["s_name"] in self.markdup_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {f['s_name']}")

            # parse sambamba output by sample name
            parsed = self.parse_markdup_stats(f)
            if parsed is not None:
                self.markdup_data[f["s_name"]] = parsed

            # filter away samples if MultiQC user does not want them
            self.markdup_data = self.ignore_samples(self.markdup_data)

            # add sambamba version to software table
            version_match = re.search(VERSION_REGEX, f["f"])
            if version_match:
                self.add_software_version(version_match.group(1), f["s_name"])

            # add results to multiqc_sources.txt
            self.add_data_source(f)

            # warn user if duplicate samples found
            if f["s_name"] in self.markdup_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")

        if len(self.markdup_data) == 0:
            return 0

        # Write parsed files to a file
        self.write_data_file(self.markdup_data, "multiqc_markdup", sort_cols=True)

        # Add sambamba markdup to general statistics table
        self.markdup_general_stats_table()

        # Add sambamba markdup bargraph section
        self.markdup_section()

        # return the length of log files parsed
        return len(self.markdup_data)

    def parse_markdup_stats(self, f):
        """
        Input actual content of sambamba markdup log output.
        Regex for important phrases and extract reads.
        Detect and calculate duplicate rate by single/paired-end samples.
        Outputs dictionary of markdup stats of a sample by single/paired-end type.
        This dict is associated with corresponding sample in key of markdup_data.

        NOTE: reads seem to sum to between 98 - 99% of input BAM file. Is Sambamba Markdup not recognizing the entire file?
        """

        regexes = {
            "sorted_end_pairs": r"sorted (\d+) end pairs",
            "single_ends": r"and (\d+) single ends",
            "single_unmatched_pairs": r"among them (\d+) unmatched",
            "duplicate_reads": r"found (\d+) duplicates",
        }
        d = {}
        for key, regex in regexes.items():
            m = re.search(regex, f["f"])
            if m:
                d[key] = int(m[1])
        if len(d) != len(regexes):
            log.debug(f"Could not find all markdup fields for '{f['fn']}' - skipping")
            return None

        # Calculate duplicate rate
        # NB: Single-end data will have 0 for sorted_end_pairs and single_unmatched_pairs
        try:
            d["duplicate_rate"] = (
                d["duplicate_reads"] / ((d["sorted_end_pairs"] * 2) + (d["single_ends"] - d["single_unmatched_pairs"]))
            ) * 100.0
        except ZeroDivisionError:
            d["duplicate_rate"] = 0
            log.debug(f"Sambamba Markdup: zero division error for '{f['fn']}'")

        # Calculate some read counts from pairs - Paired End
        if d["sorted_end_pairs"] > 0:
            d["total_sorted_paired_end_reads"] = (d["sorted_end_pairs"] * 2) - d["duplicate_reads"]
            d["total_single_end_reads"] = d["single_ends"] - (2 * d["single_unmatched_pairs"])
            d["total_single_unmatched_reads"] = 2 * d["single_unmatched_pairs"]
        # Calculate some read counts from pairs - Single End
        else:
            d["total_sorted_paired_end_reads"] = 0
            d["total_single_end_reads"] = d["single_ends"] - d["duplicate_reads"]
            d["total_single_unmatched_reads"] = 0

        return d

    def markdup_general_stats_table(self):
        """
        Take parsed stats from sambamba markdup to general stats table at the top of the report.

        Append to the dictionaries from the top-level Sambamba module script, so that all columns
        are added in a single function call across all Sambamba submodules.
        """

        self.general_stats_headers["duplicate_rate"] = {
            "title": "% Dups",
            "description": "Rate of Duplication per Sample",
            "scale": "RdYlGn-rev",
            "format": "{:,.0f}",
            "suffix": "%",
        }
        for s_name, data in self.markdup_data.items():
            if s_name not in self.general_stats_data:
                self.general_stats_data[s_name] = {}
            self.general_stats_data[s_name]["duplicate_rate"] = data["duplicate_rate"]

    def markdup_section(self):
        """
        Add markdup statistics as bar graph to multiQC report.
        """

        # plot these categories, but not duplicate rate.
        cats = {
            "total_sorted_paired_end_reads": {"name": "Total Sorted Paired End Reads"},
            "total_single_end_reads": {"name": "Total Single End Reads"},
            "total_single_unmatched_reads": {"name": "Total Single Unmatched Reads"},
            "duplicate_reads": {"name": "Total Duplicate Reads"},
        }
        config = {
            "id": "SambambaMarkdupBargraph",
            "title": "Sambamba Markdup: Duplicate Counts",
            "ylab": "# Reads",
            "cpswitch_c_active": False,
        }

        self.add_section(
            name="Sambamba Markdup",
            anchor="SambambaMarkdup",
            description="Number of reads, categorised by duplication state. **Pair counts are doubled** - see help text for details.",
            helptext="""
                The table in the Picard metrics file contains some columns referring read pairs and some referring to single reads.

                To make the numbers in this plot sum correctly, values referring to pairs are doubled according to the scheme below:

                * `Total Sorted Paired End Reads = (Sorted End Pairs * 2) - Duplicate Reads`
                * `Total Single End Reads = Single Ends - (2 * Single Unmatched Pairs)`
                * `Total Single Unmatched Reads = 2 * Single Unmatched Pairs`
            """,
            plot=bargraph.plot(self.markdup_data, cats, config),
        )



import dataclasses
import logging
from typing import Dict, List, Union, Optional
import plotly.graph_objects as go

from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset, split_long_string
from multiqc.utils import util_functions

logger = logging.getLogger(__name__)


ElemT = Union[str, float, int]


def plot(
    rows: Union[List[List[ElemT]], Dict[str, Dict[str, ElemT]]],
    pconfig: Dict,
    xcats: Optional[List[str]] = None,
    ycats: Optional[List[str]] = None,
) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    :param rows: One dataset. A dataset is a list of rows of values
    :param xcats: Labels for X axis
    :param ycats: Labels for Y axis
    :param pconfig: dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML with JS, ready to be inserted into the page
    """
    p = HeatmapPlot(rows, pconfig, xcats, ycats)

    from multiqc.utils import report

    return p.add_to_report(report)


class HeatmapPlot(Plot):
    @dataclasses.dataclass
    class Dataset(BaseDataset):
        rows: List[List[ElemT]]
        xcats: List[str]
        ycats: List[str]

        @staticmethod
        def create(
            dataset: BaseDataset,
            rows: Union[List[List[ElemT]], Dict[str, Dict[str, ElemT]]],
            xcats: Optional[List[str]] = None,
            ycats: Optional[List[str]] = None,
        ) -> "HeatmapPlot.Dataset":
            if isinstance(rows, dict):
                # Convert dict to a list of lists
                if not ycats:
                    ycats = list(rows.keys())
                if not xcats:
                    xcats = []
                    for y, value_by_x in rows.items():
                        for x, value in value_by_x.items():
                            if x not in xcats:
                                xcats.append(x)
                rows = [[rows.get(y, {}).get(x) for x in xcats] for y in ycats]

            dataset = HeatmapPlot.Dataset(
                **dataset.__dict__,
                rows=rows,
                xcats=xcats,
                ycats=ycats,
            )
            return dataset

        def create_figure(
            self,
            layout: Optional[go.Layout] = None,
            is_log=False,
            is_pct=False,
        ) -> go.Figure:
            """
            Create a Plotly figure for a dataset
            """
            return go.Figure(
                data=go.Heatmap(
                    z=self.rows,
                    x=self.xcats,
                    y=self.ycats,
                    **self.trace_params,
                ),
                layout=layout or self.layout,
            )

    def __init__(
        self,
        rows: Union[List[List[ElemT]], Dict[str, Dict[str, ElemT]]],
        pconfig: Dict,
        xcats: Optional[List[str]],
        ycats: Optional[List[str]],
    ):
        super().__init__(PlotType.HEATMAP, pconfig, n_datasets=1)

        if isinstance(rows, list):
            if ycats and not isinstance(ycats, list):
                raise ValueError(
                    f"Heatmap plot {self.id}: ycats must be passed as a list when the input data is a 2d list. "
                    f"The order of that list should match the order of the rows in the input data."
                )
            if xcats and not isinstance(xcats, list):
                raise ValueError(
                    f"Heatmap plot {self.id}: xcats must be passed as a list when the input data is a 2d list. "
                    f"The order of that list should match the order of the columns in the input data."
                )

        self.layout.update(
            yaxis=dict(
                # Prevent JavaScript from automatically parsing categorical values as numbers:
                type="category",
            ),
            xaxis=dict(
                # Prevent JavaScript from automatically parsing categorical values as numbers:
                type="category",
            ),
            showlegend=pconfig.get("legend", True),
        )

        self.square = pconfig.get("square", True)  # Keep heatmap cells square

        # Extend each dataset object with a list of samples
        self.datasets: List[HeatmapPlot.Dataset] = [
            HeatmapPlot.Dataset.create(
                self.datasets[0],
                rows=rows,
                xcats=xcats,
                ycats=ycats,
            )
        ]

        self.min = self.pconfig.get("min", None)
        if self.min is None:
            for dataset in self.datasets:
                for row in dataset.rows:
                    for val in row:
                        if val is not None and isinstance(val, (int, float)):
                            self.min = val if self.min is None else min(self.min, val)
        self.max = self.pconfig.get("max", None)
        if self.max is None:
            for dataset in self.datasets:
                for row in dataset.rows:
                    for val in row:
                        if val is not None and isinstance(val, (int, float)):
                            self.max = val if self.max is None else max(self.max, val)

        # Determining the size of the plot to reasonably display data without cluttering it too much.
        # For flat plots, we try to make the image large enough to display all samples, but to a limit
        # For interactive plots, we set a lower default height, as it will possible to resize the plot
        num_rows = len(self.datasets[0].ycats)
        num_cols = len(self.datasets[0].xcats)
        MAX_HEIGHT = 900 if self.flat else 500  # smaller number for interactive, as it's resizable
        MAX_WIDTH = 900  # default interactive width can be bigger

        # Number of samples to the desired size in pixel one sample will take on a screen
        def n_elements_to_size(n: int):
            if n >= 50:
                return 11
            if n >= 30:
                return 18
            if n >= 20:
                return 26
            if n >= 15:
                return 33
            if n >= 10:
                return 45
            if n >= 5:
                return 60
            if n >= 3:
                return 80
            if n >= 2:
                return 100
            return 120

        x_px_per_elem = n_elements_to_size(num_cols)
        y_px_per_elem = n_elements_to_size(num_rows)
        min_px_per_elem = min(x_px_per_elem, y_px_per_elem)
        if self.square:
            x_px_per_elem = y_px_per_elem = min_px_per_elem

        width = pconfig.get("width") or int(num_cols * x_px_per_elem)
        height = pconfig.get("height") or int(num_rows * y_px_per_elem)

        if not self.square and width < MAX_WIDTH and x_px_per_elem < 40:  # can fit more columns on the screen
            logger.debug(f"Resizing width from {width} to {MAX_WIDTH} to fit horizontal column text on the screen")
            width = MAX_WIDTH
            x_px_per_elem = width / num_cols

        if height > MAX_HEIGHT or width > MAX_WIDTH:
            logger.debug(f"Resizing from {width}x{height} to fit the maximum size {MAX_WIDTH}x{MAX_HEIGHT}")
            if self.square:
                px_per_elem = min(MAX_WIDTH / num_cols, MAX_HEIGHT / num_rows)
                width = height = int(num_rows * px_per_elem)
            else:
                x_px_per_elem = MAX_WIDTH / num_cols
                y_px_per_elem = MAX_HEIGHT / num_rows
                width = int(num_cols * x_px_per_elem)
                height = int(num_rows * y_px_per_elem)

        logger.debug(f"Heatmap size: {width}x{height}, px per element: {x_px_per_elem:.2f}x{y_px_per_elem:.2f}")

        # For not very large datasets, making sure all ticks are displayed:
        if y_px_per_elem > 12:
            self.layout.yaxis.tickmode = "array"
            self.layout.yaxis.tickvals = list(range(num_rows))
            self.layout.yaxis.ticktext = ycats
        if x_px_per_elem > 18:
            self.layout.xaxis.tickmode = "array"
            self.layout.xaxis.tickvals = list(range(num_cols))
            self.layout.xaxis.ticktext = xcats
        if pconfig.get("angled_xticks", True) is False and x_px_per_elem >= 40:
            # Break up the horizontal ticks by whitespace to make them fit better vertically:
            self.layout.xaxis.ticktext = ["<br>".join(split_long_string(cat, 10)) for cat in xcats]
            # And leave x ticks horizontal:
            self.layout.xaxis.tickangle = 0
        else:
            # Rotate x-ticks to fit more of them on screen
            self.layout.xaxis.tickangle = 45

        self.layout.height = 200 + height
        self.layout.width = (250 + width) if self.square else None

        self.layout.xaxis.showgrid = False
        self.layout.yaxis.showgrid = False
        self.layout.yaxis.autorange = "reversed"  # to make sure the first sample is at the top
        self.layout.yaxis.ticklabelposition = "outside right"

        colstops = pconfig.get("colstops", None)
        if colstops:
            # A list of 2-element lists where the first element is the
            # normalized color level value (starting at 0 and ending at 1),
            # and the second item is a valid color string.
            try:
                colorscale = [[float(x), color] for x, color in colstops]
            except ValueError:
                colorscale = None
            else:
                # normalise the stop to a range from 0 to 1
                minval = min(x for x, _ in colorscale)
                maxval = max(x for x, _ in colorscale)
                rng = maxval - minval
                colorscale = [[(x - minval) / rng, color] for x, color in colorscale]
        else:
            # default colstops
            colorscale = [
                [0, "#313695"],
                [0.1, "#4575b4"],
                [0.2, "#74add1"],
                [0.3, "#abd9e9"],
                [0.4, "#e0f3f8"],
                [0.5, "#ffffbf"],
                [0.6, "#fee090"],
                [0.7, "#fdae61"],
                [0.8, "#f46d43"],
                [0.9, "#d73027"],
                [1, "#a50026"],
            ]

        decimal_places = pconfig.get("decimalPlaces", 2)

        xlab = pconfig.get("xlab", "x")
        ylab = pconfig.get("ylab", "y")
        zlab = pconfig.get("zlab", "z")
        hovertemplate = f"{xlab}: %{{x}}<br>{ylab}: %{{y}}<br>{zlab}: %{{z}}<extra></extra>"

        for ds in self.datasets:
            ds.trace_params = {
                "colorscale": colorscale,
                "reversescale": pconfig.get("reverseColors", False),
                "showscale": pconfig.get("legend", True),
                "zmin": pconfig.get("min", None),
                "zmax": pconfig.get("max", None),
                "hovertemplate": hovertemplate,
            }
            # Enable datalabels if there are less than 20x20 cells, unless heatmap_config.datalabels is set explicitly
            if pconfig.get("datalabels") is None and num_rows * num_cols < 400:
                ds.trace_params["texttemplate"] = "%{z:." + str(decimal_places) + "f}"

    def dump_for_javascript(self) -> Dict:
        """Serialise the plot data to pick up in JavaScript"""
        d = super().dump_for_javascript()
        d["xcats_samples"] = self.pconfig.get("xcats_samples", True)
        d["ycats_samples"] = self.pconfig.get("ycats_samples", True)
        d["square"] = self.square
        return d

    def buttons(self) -> List[str]:
        """
        Heatmap-specific controls, only for the interactive version.
        """
        buttons = super().buttons()

        if not self.flat:
            # find min val across all datasets across all cols and rows
            buttons.append(
                f"""
            <div class="mqc_hcplot_range_sliders">
                <div>
                    <label for="{self.id}_range_slider_min_txt">Min:</label>
                    <input id="{self.id}_range_slider_min_txt" type="number" class="form-control" 
                        value="{self.min}" data-target="{self.id}" data-minmax="min" min="{self.min}" max="{self.max}" />
                    <input id="{self.id}_range_slider_min" type="range" 
                        value="{self.min}" data-target="{self.id}" data-minmax="min" min="{self.min}" max="{self.max}" step="any" />
                </div>
                <div style="margin-left: 30px;">
                    <label for="{self.id}_range_slider_max_txt">Max:</label>
                    <input id="{self.id}_range_slider_max_txt" type="number" class="form-control" 
                        value="{self.max}" data-target="{self.id}" data-minmax="max" min="{self.min}" max="{self.max}" />
                    <input id="{self.id}_range_slider_max" type="range" 
                        value="{self.max}" data-target="{self.id}" data-minmax="max" min="{self.min}" max="{self.max}" step="any" />
                </div>
            </div>
            """
            )
        return buttons

    def save_data_file(self, dataset: Dataset) -> None:
        data = [
            ["."] + dataset.xcats,
        ]
        for ycat, row in zip(dataset.ycats, dataset.rows):
            data.append([ycat] + row)

        util_functions.write_data_file(data, dataset.uid)
